/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: LTTUnit.m,v 1.17 1996/07/03 14:33:40 tiggr Exp $  */

#import "ltt.h"

/* All units currently known.  */
TLDictionary *ltt_units;

/* All instances known in all known units.  */
static TLDictionary *ltt_instances;

/* Used when searching in the hierarchy.  */
static int search_mark;

@implementation LTTUnit

+(LTTUnit *) findUnitNamed: (TLString *) n
{
  LTTUnit *u = [self unitNamed: n];

  if (!u)
    {
      [self searchUnit: n alongPath: load_path];
      u = [self unitNamed: n];
    }

  return u;
} /* +findUnitNamed: */

+initialize
{
  if (!ltt_units)
    {
      ltt_units = [TLDictionary dictionary];
      [ltt_units gcLock];
      ltt_instances = [TLDictionary dictionary];
      [ltt_instances gcLock];
    }

  return self;
} /* +initialize */

+(TLVector *) instancesNamed: (id <TLString>) n
{
  return [ltt_instances objectForKey: n];
} /* +instancesNamed: */

+(LTTUnit *) searchUnit: (TLString *) n alongPath: (TLVector *) path
{
  TLString *filename;
  LTTUnit *u;

  if (load_unit (n, path, (id) self, &filename))
    return nil;

  u = [ltt_units objectForKey: n];
  if (!u)
    error (@"`%@' fails to define the unit `%@'", filename, n);

  return u;
} /* +searchUnit:alongPath: */

+(id <LTTUnitAcceptor>) startUnitNamed: (id <TLString>) n
{
  TLString *s = [current_filename string];

  /* The directory must be deduced from the current filename.  */
  id <TLRange> r = [s rangeOfString: @"/" options: TLSEARCH_BACKWARD];
  TLString *d = (r ? [s stringWithRange:
			[TLRange rangeWithStart: 0 end: [r _start]]]
		 : (id) @".");

  return [self unitWithName: n directory: d];
} /* +startUnitNamed: */

+(LTTUnit *) unitNamed: (TLString *) n
{
  return [ltt_units objectForKey: n];
} /* +unitNamed: */

+(LTTUnit *) unitWithName: (TLString *) n directory: (TLString *) d
{
  LTTUnit *u = [[self gcAlloc] initWithName: n directory: d];

  [ltt_units setObject: u forKey: n];

  return u;
} /* +unitWithName:directory: */

-(void) addExtension: (LTTExtension *) e
{
  [extensions addElement: e];
} /* -addExtension: */

-(void) addInstance: (LTTInstance *) m
{
  TLString *s = [[m name] internal];
  TLVector *v;

  /* Add the instance to ourself.  */
  [instances setObject: m forKey: s];

  /* Add the instance to the dict of all instances.  */
  v = [ltt_instances objectForKey: s];
  if (!v)
    {
      v = [TLVector vector];
      [ltt_instances setObject: v forKey: s];
    }
  [v addElement: m];
}

-(BOOL) dependsOn: (LTTUnit *) unit
	     mark: (int) k
{
  if (mark == k)
    return NO;
  mark = k;

  if (self == unit)
    return YES;

  {
    int i, n = [units_used length];

    for (i = 0; i < n; i++)
      {
	LTTUnit *u = [units_used _elementAtIndex: i];

	if (u == unit || [u dependsOn: unit mark: k])
	  return YES;
      }

    return NO;
  }
}

-(BOOL) dependsOn: (LTTUnit *) u
{
  return [self dependsOn: u mark: ++search_mark];
}

-(TLString *) directory
{
  return dir;
} /* -directory */

-(id <TLEnumerator>) extensions
{
  return [extensions enumerator];
}

-(id <TLEnumerator>) files
{
  return [files valueEnumerator];
} /* -files */

-(void) gcReference
{
  MARK (dir);
  MARK (files);
  MARK (instances);
  MARK (extensions);
  MARK (units_used);

  [super gcReference];
}

-(id) initWithName: (TLString *) n directory: (TLString *) d
{
  if (![super initWithName: [LTTName nameWithInternal: n]])
    return nil;

  dir = d;
  files = [TLDictionary dictionary];
  instances = [TLDictionary dictionary];
  extensions = [TLVector vector];
  units_used = [TLVector vector];

  return self;
}

-(LTTInstance *) instanceNamed: (TLString *) s
{
  return [instances objectForKey: s];
} /* -instanceNamed: */

-(id <TLEnumerator>) instances
{
  return [instances valueEnumerator];
} /* -instances */

-(LTTInstance *) metaWithName: (TLString *) s
{
  LTTInstance *m = [self instanceNamed: s];

  if (m)
    error (@"attempt to add duplicate class `%@' to unit `%@'", s, name);
  else
    {
      m = [LTTMeta instanceAndClassWithName: s unit: self];
      [self addInstance: m];
    }

  return m;
} /* -metaWithName: */

-(LTTFile *) fileNamed: (TLString *) s
{
  return [files objectForKey: s];
} /* -fileNamed: */

-(TLCons *) orderedUnits: (TLCons *) l
		    mark: (int) k
{
  int i, n;

  if (mark == k)
    return l;
  mark = k;

  if (units_used)
    for (i = 0, n = [units_used length]; i < n; i++)
      l = [[units_used _elementAtIndex: i] orderedUnits: l mark: k];

  return l ? [l nconc: CONS (self, nil)] : CONS (self, nil);
}

-(TLCons *) orderedUnits: (TLCons *) l
{
  return [self orderedUnits: l mark: ++search_mark];
}

-(id <TLString>) outputDefinitionName
{
  return (id) formac (nil, @"%@%@", TO_UNIT_DEF_PREFIX, [name external]);
}

-(void) resolvePosing
{
  while (posing)
    {
      TLString *meta_name, *unit_name = nil;
      LTTExtension *x;
      TLCons *entry;
      LTTMeta *m;

      DECONS (posing, entry, posing);
      DECONS (entry, x, entry);

      while (entry)
	{
	  DECONS (entry, meta_name, entry);

	  if ([meta_name consp])
	    DECONS ((id) meta_name, unit_name, meta_name);

	  m = ltt_instance_in_unit (unit_name, self,
				    [isa instancesNamed: meta_name]);
	  if (m)
	    {
	      if ([[x meta] poser])
		warning (@"%@.%@ posing %@.%@ actually poses %@.%@",
			 [[m unit] name], [m name], [[[x meta] unit] name],
			 [[x meta] name], [[[[x meta] posedSelf] unit] name],
			 [[[x meta] posedSelf] name]);

	      [m setPoser: [x meta]];
	      [[m itsClass] setPoser: [[x meta] itsClass]];
	    }
	}
    }
} /* -resolvePosing */

-(TLVector *) unitsUsed
{
  return units_used;
}

/******************** LTTUnitAcceptor protocol ********************/

-(id <LTTUnitAcceptor>) class: (id <TLString>) c
		       inFile: (id <TLString>) f
		       posing: (TLCons *) p
{
  return [self extension: nil inFile: f ofClass: c
	       inUnit: [name internal] posing: p];
} /* -class:inFile:posing: */

-(id <LTTUnitAcceptor>) endUnit
{
  return isa;
} /* -endUnit */

-(id <LTTUnitAcceptor>) extension: (id <TLString>) ext_name
			   inFile: (id <TLString>) filename
			  ofClass: (id <TLString>) classname
			   inUnit: (id <TLString>) unit_name
			   posing: (TLCons *) p
{
  LTTFile *f;
  LTTMeta *m;
  LTTUnit *u;

  u = unit_name ? [LTTUnit unitNamed: unit_name] : self;
  if (!u)
    {
      error (@"unknown unit `%@'", unit_name);
      return self;
    }

  if (!filename)
    filename = ext_name ?: classname;

  f = [self fileNamed: filename];
  if (!f)
    {
      TLString *full_name = (id) formac (nil, @"%@/%@", dir, filename);
      f = [LTTFile fileWithName: full_name unit: self];
      [files setObject: f forKey: filename];
    }

  m = ltt_instance_in_unit (unit_name, self,
			    [isa instancesNamed: classname]);
  if (m && !ext_name && [m extensionNamed: nil])
    {
      error (@"attempt to redefine class `%@' in `%@'", classname, filename);
      cerror (@"class `%@' already defined in `%@'", classname,
	      [[[m extensionNamed: nil] container] name]);
    }
  else if (m && ext_name && [m extensionNamed: ext_name])
    {
      error (@"attempt to redefine %@(%@) in `%@'",
	     classname, ext_name, filename);
      cerror (@"%@(%@) already defined in `%@'", classname, ext_name,
	      [[[m extensionNamed: nil] container] name]);
    }
  else
    {
      LTTExtension *x;

      if (!m)
	m = [self metaWithName: classname];

      x = [LTTExtension extensionWithName: ext_name file: f meta: m];
      [LTTExtension extensionWithName: ext_name file: f meta: [m itsClass]];

      if (p)
	posing = CONS (CONS (x, p), posing);
    }

  return self;
} /* -extension:inFile:ofClass:inUnit: */

-(id <LTTUnitAcceptor>) usesUnit: (id <TLString>) n
{
  LTTUnit *u = [isa findUnitNamed: n];

  if (u)
    {
      [units_used addElement: u];
      if ([u dependsOn: self])
	error (@"circular dependency between units `%@' and `%@'", name, n);
    }

  return self;
}

@end
