/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: OTMCompound.m,v 1.30 1996/06/07 11:39:28 tiggr Exp $  */

#import "OTMCompound.h"
#import "OTMBasic.h"
#import "OTMCustomMethod.h"
#import "OTMVariable.h"
#import "OTMLocalVar.h"

@implementation OTMCompound

+(OTMCompound *) compoundWithContainer: (id) c
{
  return [[self gcAlloc] initWithContainer: c];
} /* +compoundInContainer: */

-(void) addStatement: (id) stmt
{
  stmt = [stmt precompile];

  if (stmt)
    [stats addElement: stmt];
} /* -addStatement: */

-(void) addVariable: (OTMVariable *) v
{
  id prev = nil;

  if (!vars)
    ASGN_IVAR (vars, [TLDictionary dictionary]);
  else
    {
      prev = [vars objectForKey: [v name]];
      if (prev)
	{
	  error_for (v, @"redeclaration of variable `%@'", [v name]);
	  cerror_for (prev, @"previous declaration was here");
	}
    }

  if (!prev)
    [vars setObject: v forKey: [v name]];
} /* -addVariable */

-(void) assignResult: (OTMExpr *) v
{
  if (!v)
    abort ();

  if (!result && [v type] == basic_type[BT_VOID])
    result = void_expr;
  else if (result != void_expr)
    {
      if (!result)
	{
	  id previous_compound = current_compound;

	  current_compound = container;
	  result = temp_something_with_type ([v type], YES);
	  current_compound = previous_compound;
	}

      emit_assignment (result, v);
    }
} /* -assignResult: */

-(void) compile
{
#if STACK_REF_STRUCT
  TLVector *local_refs = nil;
#endif
  OTMVariable *v;
  
  formac (of, @"%@{", nl ());
  indent_level++;

  /* Output local variable declarations.  */
  if (vars)
    {
      id <TLEnumerator> e = [vars valueEnumerator];

      while ((v = [e nextObject]))
#if STACK_REF_STRUCT
	if ([[v type] isObjectType])
	  {
	    if (!local_refs)
	      local_refs = [TLVector vector];
	    [local_refs addElement: v];
	  }
	else
#endif
	  {
	    formac (of, @"%@", nl ());
	    [v outputDeclaration: of];
	    formac (of, @";");
	  }
    }

#if STACK_REF_STRUCT
  /* Output the stack protection.  */
  if ([container isKindOf: [OTMMethod class]]
      && ([(OTMCustomMethod *) container referencingArguments]
	  || local_refs))
    {
      OTMCustomMethod *m = (OTMCustomMethod *) container;
      TLVector *args = [m referencingArguments];
      int num_locals = [local_refs length];
      int num_args = [args length];
      int i;

      [m setStackProtected];

      if (num_args)
	formac (of, @"void **%@[%d];\n", TO_PROTECT_STACK_ARGS, num_args);

      formac (of, @"struct {\nstruct trt_protect_stack _c;\n");

      for (i = 0; i < num_locals; i++)
	{
	  v = [local_refs _elementAtIndex: i];
	  [v outputDeclaration: of];
	  formac (of, @";\n");
	}

      formac (of, @"} %@;\n", TO_PROTECT_STACK);

      for (i = 0; i < num_args; i++)
	{
	  v = [args _elementAtIndex: i];
	  formac (of, @"%@[%d] = (void **) &%@;\n", TO_PROTECT_STACK_ARGS, i,
		  [v outputReference]);
	}

      formac (of, @"%@._c.next = %@;\n",
	      TO_PROTECT_STACK, TO_TRT_PROTECT_STACK);
      formac (of, @"%@._c.num_args = %d;\n", TO_PROTECT_STACK, num_args);
      formac (of, @"%@._c.num_locals = %d;\n", TO_PROTECT_STACK, num_locals);
      formac (of, @"%@._c.arguments = %@;\n", TO_PROTECT_STACK,
	      num_args ? TO_PROTECT_STACK_ARGS : @"0");

      formac (of, @"%@ = &%@._c;\n", TO_TRT_PROTECT_STACK, TO_PROTECT_STACK);

      if (num_locals)
	{
	  v = [local_refs _elementAtIndex: 0];
	  EFILLZERO ([v outputReference], num_locals);
	}

      formac (of, @"\n");
    }
#endif

#if CACHE_EXTENSION_POINTER
  /* If we're a method's body, tell the method to output the
     initializations of the extension pointers.  */
  if ([container isKindOf: [OTMMethod class]])
    [(OTMCustomMethod *) container compileExtensionPointerDefinitions];
#endif

  /* Compile the actual body.  */
  {
    id <TLEnumerator> e = [stats enumerator];
    id st;

    [self compileCondition: 0];

    while ((st = [e nextObject]) && !this_num_errors)
      [st compileStatement];

    [self compileCondition: 1];
  }

  /* Output the method epilogue.  */
  if ((id) container == output_current_method)
    {
      OTMExpr *e;

      if ([output_current_method haveReturnStatement])
	formac (of, @"\n%@:", TO_RETURN_LABEL);

#if STACK_REF_STRUCT
      {
	BOOL protected = [output_current_method stackProtected];

	if (protected)
	  formac (of, @"%@%@ = %@._c.next; ", nl (),
		  TO_TRT_PROTECT_STACK, TO_PROTECT_STACK);
      }
#endif

      e = [output_current_method returnValue];
      if (e)
	{
	  if ([e isTuple])
	    [output_current_method assignOutArgsFrom: (id) e];
	  formac (of, @"%@return %@;", nl (), [e result]);
	}
      else if ([output_current_method returnType] != basic_type[BT_VOID])
	error (@"return value never assigned in method returning non-void");
    }

  indent_level--;
  formac (of, @"%@}", nl ());
} /* -compile */

-(void) compileAssignmentToTuple: (OTMTuple *) lhs
{
  if (!value)
    error_for (self, @"compound has no value");
  else
    [value compileAssignmentToTuple: lhs];
} /* -compileAssignmentToTuple: */

-(void) compileCondition: (BOOL) end_not_start
{
} /* -compileCondition: */

-(OTMCompound *) container
{
  return container;
} /* -container */

-(id) elaborate
{
  /* If the VALUE has been set, make sure it exists in the enclosing
     compound.  */
  if (value)
    {
      id previous_compound = current_compound;

      current_compound = self;
      [self assignResult: value];
      current_compound = previous_compound;
    }

  return self;
} /* -elaborate */

-(void) gcReference
{
  MARK (value);
  MARK (result);
  MARK (container);
  MARK (stats);
  MARK (vars);
  MARK (temp_free);
  MARK (temp_use);

  [super gcReference];
} /* -gcReference */

-initWithContainer: (id) c
{
  if (![super init])
    return nil;

  container = c;
  stats = [TLVector vector];

  return self;
} /* -initWithContainer: */

-(void) releaseTemporaryVariables
{
  if (temp_use)
    {
      if (!temp_free)
	temp_free = temp_use;
      else
	[temp_free addElementsFromSequence: temp_use];
      temp_use = nil;
    }
} /* -releaseTemporaryVariables */

-(TLCons *) resolveWithExpected: (TLCons *) expected
		    convertible: (OTMType *) to
			context: (OTMType *) cxt
			indices: (int *) indices
			  index: (int) index
{
  if (!type)
    {
      if (!value)
	{
	  error_for (self, @"compound without value");
	  return nil;
	}
      else
	type = [value type];
    }

  return [super resolveWithExpected: expected convertible: to
		context: cxt indices: indices index: index];
}

-(id <TLString>) result
{
  if (!result)
    ABORT ();

  return [result result];
} /* -result */

-(OTMVariable *) searchVariableNamed: (id <TLString>) n
{
  if (vars)
    {
      id v = [vars objectForKey: n];
      if (v)
	return v;
    }
  return [container searchVariableNamed: n];
} /* -searchVariableNamed: */

-(void) setBreakValue: (OTMExpr *) e
{
  [container setBreakValue: e];
} /* -setBreakValue: */

-(void) setValue: (OTMExpr *) v
{
  ASGN_IVAR (value, v);
} /* -setValue: */

-(OTMVariable *) temporaryWithType: (OTMType *) t
{
  OTMVariable *v = nil;

  if (temp_free)
    {
      int i, n = [temp_free length];

      for (i = 0; i < n; i++)
	{
	  OTMVariable *o = [temp_free _elementAtIndex: i];

	  if ([o type] == t)
	    {
	      [temp_free _removeElementAtIndex: i];

	      v = o;
	      break;
	    }
	}
    }

  if (!v)
    {
      v = [OTMLocalVar temporaryVariableWithType: t];
      emit_local_var (v);
    }

  if (!temp_use)
    temp_use = [TLVector vector];

  [temp_use addElement: v];

  return v;
} /* -temporaryWithType: */

-(OTMExpr *) value
{
  return result;
}

@end
