/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: OTMExtension.m,v 1.46 1996/08/27 10:05:52 tiggr Exp $  */

#define OTMEXTENSION_DECLARE_PRIVATE_METHODS
#import "OTMExtension.h"
#import "OTMAlias.h"
#import "OTMClass.h"
#import "OTMConstant.h"
#import "OTMCustomMethod.h"
#import "OTMMeta.h"
#import "OTMObjectVar.h"
#import "global.h"

@implementation LTTExtension (Semantics)

+(id) semanticsForExtension: (LTTExtension *) ext
{
  return [[OTMExtension gcAlloc] initWithStructure: ext];
} /* +semanticsForExtension: */

@end

@implementation OTMExtension

-(void) addMethod: (OTMMethod *) m
{
  TLVector *v = [self methodsNamed: [m firstNamePart] create: YES];

  [v addElement: m];
} /* -addMethod: */

-(OTMObjectVar *) aliasWithName: (id <TLString>) n
			   type: (id) t
		     qualifiers: (otm_qualifiers) q
{
  OTMObjectVar *var;
  OTMMeta *meta;

  if (![self cvDeclarationOK: n])
    return nil;

  meta = [[structure meta] semantics];
  var = [meta searchEntityNamed: n supers: YES class: NO
	      variables: YES constants: NO];
  if (!var)
    error (@"no variable named %@ exists", n);
  else
    {
      OTMType *ot = [var type];

      if (![t matches: ot])
	error (@"inapplicable alias type: %@ does not match %@",
	       type_name (t), type_name (ot));
      else if ([ot matchesConvertibly: t])
	error (@"inapplicable alias type: %@ does not fit %@",
	       type_name (t), type_name (ot));
      else
	{
	  if (!aliases)
	    aliases = [TLDictionary dictionary];
	  var = (id) [OTMAlias aliasWithExtension: self variable: var
			       type: t qualifiers: q];
	  [aliases setObject: var forKey: n];
	}
    }

  return var;
} /* -aliasWithName:type:qualifiers: */

-(void) checkImplementation
{
  BOOL is_main_ext = structure == [[structure meta] extensionNamed: nil];
  BOOL is_deferred = [[[structure meta] semantics] deferredp];
  id <TLEnumerator> ev = [methods valueEnumerator];
  BOOL have_moaned = NO;
  TLVector *v;

  while ((v = [ev nextObject]))
    {
      id <TLEnumerator> em = [v enumerator];
      OTMMethod *m;

      while ((m = [em nextObject]))
	{
	  if ([m declaredp] && ![m definedp])
	    if ([m deferredp])
	      {
		if (!is_deferred)
		  error_for (m, @"deferred method `%@' in undeferred class",
			     [m name]);
	      }
	    else
	      {
		if (!have_moaned)
		  {
		    if (is_main_ext)
		      error (@"incomplete implementation of %@",
			     ltt_meta_name ([structure meta]));
		    else
		      error (@"incomplete implementation of %@",
			     ltt_ext_name (structure));
		    have_moaned = YES;
		  }
		error_for (m, @"method `%@' declared undeferred "
			   @"but not defined", [m name]);
	      }
	}
    }

  /* XXX Do many more checks with respect to super classes.  */
} /* -checkImplementation */

-(void) compileDeclareExtensionIdentityName
{
  if (!declared_eid)
    {
      declared_eid = YES;
      formac (of, @"extern %@ %@;\n", TO_EXT_ID_TYPE,
	      [structure outputExtensionIdentityName]);
    }
} /* -compileDeclareExtensionIdentityName */

-(void) compileObjVarTypeDeclarations
{
  if (!declared_ivar_struct && (vars_order || vars_static))
    {
      id <TLEnumerator> e;

      declared_ivar_struct = YES;

      if (vars_static)
	{
	  OTMObjectVar *v;

	  e = [vars_static valueEnumerator];
	  while ((v = [e nextObject]))
	    {
	      [[v type] compileDeclaration];
	      [v compileDeclaration];
	    }
	}

      if (vars_order)
	{
	  OTMVariable *v;

	  e = [vars_order enumerator];
	  while ((v = [e nextObject]))
	    [[v type] compileDeclaration];

	  formac (of, @"struct %@ %@;\n", [structure outputExtensionStructName],
		  [self outputDeclaration]);
	}
    }
} /* -compileObjVarTypeDeclarations */

-(OTMConstant *) constantNamed: (id <TLString>) n
{
  return constants ? [constants objectForKey: n] : nil;
} /* -constantNamed: */

-(OTMConstant *) constantWithName: (id <TLString>) n
		       qualifiers: (otm_qualifiers) q
			    value: (OTMExpr *) val
{
  OTMConstant *cst;

  /* XXX Handle qualifiers.  */

  if (![self cvDeclarationOK: n])
    return nil;

  cst = [OTMConstant constantWithExtension: self name: n value: val];
  if (!constants)
    constants = [TLDictionary dictionary];
  [constants setObject: cst forKey: n];

  return cst;
} /* -constantWithName:type:qualifiers: */

-(BOOL) cvDeclarationOK: (id <TLString>) n
{
  OTMMeta *whole = [[structure meta] semantics];
  OTMObjectVar *v;

  /* Allow a subclass to define a new variable with the same name.  */
  v = [whole entityNamed: n variables: YES constants: YES];
  if (v)
    {
      error (@"duplicate definition of %@", n);
      cerror_for (v, @"%@ previously defined in %@", n,
		  ltt_ext_name ([[v extension] structure]));
      return NO;
    }

  /* Do not check for clashes between the State instance and State class.  */
  if (structure != ltt_ext_i_state)
    {
      v = [[[[whole structure] itsClass] semantics]
		    searchEntityNamed: n supers: YES class: NO
		    variables: YES constants: YES];
      if (v)
	{
	  warning (@"definition of %@ shadows previous definition", n);
	  cwarning_for (v, @"%@ previously defined in %@", n,
			ltt_ext_name ([[v extension] structure]));
	}
    }

  return YES;
} /* -cvDeclarationOK: */

-(void) description: (id <TLMutableStream>) stream
{
  [super description: stream];

  formac (stream, @" %@ %@",
	  [[structure name] internal], ltt_meta_name ([structure meta]));
}

-(void) dumpInfo: (id <TLOutputStream>) s
{
  OTMMeta *basis = [[structure meta] semantics];
  id previous_context = output_current_context;
  id <TLString> name = [[structure name] internal];
  id <TLEnumerator> e;
  BOOL printed = NO;
  TLVector *vec;

  output_current_context = [structure meta];

  formac (s, @"\n(extension %# %@ (file %#)",
	  name, [[[structure meta] semantics] typeInfo],
	  [[[structure container] name] internal]);

  if (vars_order || vars_static)
    {
      OTMObjectVar *v;

      formac (s, @"\n (variables %d %#", vars_align,
	      vars_order ? [self outputDeclaration] : nil);

      e = [vars_order enumerator];
      while ((v = [e nextObject]))
	[v dumpInfo: s];

      e = [vars_static valueEnumerator];
      while ((v = [e nextObject]))
	[v dumpInfo: s];

      [@")" print: s quoted: NO];
    }

  if (constants)
    {
      OTMConstant *c;

      formac (s, @"\n (constants");

      e = [constants valueEnumerator];
      while ((c = [e nextObject]))
	[c dumpInfo: s];

      [@")" print: s quoted: NO];
    }

  [basis dumpInfo: s];

  e = [methods valueEnumerator];
  while ((vec = [e nextObject]))
    {
      int i, n = [vec length];

      if (!printed)
	{
	  formac (s, @"\n (methods");
	  printed = YES;
	}

      for (i = 0; i < n; i++)
	{
	  OTMMethod *m = [vec _elementAtIndex: i];

	  [m dumpInfo: s];
	}
    }

  if (printed)
    formac (s, @")");

  formac (s, @")");

  output_current_context = previous_context;
}

-(void) endCompile
{
  formac (of, @"/* End of %@.  */\n", ltt_ext_name (structure));
} /* -endCompile */

-(TLCons *) methodsNamed: (TLVector *) name_parts
		  sender: (OTMMeta *) sender
		    list: (TLCons *) l
{
  TLVector *v = [self methodsNamed: [name_parts _elementAtIndex: 0] create: NO];
  OTMMeta *sender_other = ([sender classp] ? (id) [(OTMClass *) sender instance]
			   : [sender itsClass]);
  LTTMeta *so_str = [sender_other structure];
  LTTMeta *sender_str = [sender structure];
  LTTMeta *str = [structure meta];  
  OTMMeta *sem = [str semantics];

  if (v)
    {
      id <TLEnumerator> e = [v enumerator];
      OTMMethod *m;

      while ((m = [e nextObject]))
	if ([m namePartsMatch: name_parts])
	  {
	    if (sender)
	      {
		otm_qualifiers q = [m protection];

		if (q && !(q == OQ_PUBLIC
			   || (q == OQ_PRIVATE
			       && (sender == sem || sender_other == sem))
			   || (q == OQ_PROTECTED
			       && (sender == sem || sender_other == sem
				   || [str isProperSub: sender_str]
				   || [str isProperSub: so_str]))))
		  continue;
	      }

	    l = CONS (m, l);
	  }
    }

  return l;
}

-(void) resolveIdentifiers: (OTMMeta *) meta
{
  id <TLEnumerator> e = [methods valueEnumerator];
  TLVector *v;

  while ((v = [e nextObject]))
    {
      id <TLEnumerator> e = [v enumerator];
      OTMMethod *m;

      while ((m = [e nextObject]))
	[m resolveIdentifiers: meta];
    }
}

-(OTMMethod *) searchSimilarMethod: (OTMMethod *) method
{
  TLVector *v = [self methodsNamed: [method firstNamePart] create: NO];
  id <TLEnumerator> e;
  OTMMethod *m;

  if (!v)
    return nil;

  e = [v enumerator];
  while ((m = [e nextObject]))
    {
      if ([(id) [method name] equal: [m name]] && [method typesMatch: m])
	return m;
    }

  return nil;
} /* -searchSimilarMethod: */

-(void) gcReference
{
  MARK (structure);
  MARK (methods);
  MARK (vars_keyed);
  MARK (vars_order);
  MARK (vars_static);
  MARK (aliases);
  MARK (constants);

  [super gcReference];
} /* -gcReference */

-(BOOL) hasVariables
{
  return !!vars_order;
} /* -hasVariables */

-initWithStructure: (LTTExtension *) str
{
  if (![super init])
    return nil;

  structure = str;
  methods = [TLDictionary dictionary];

  return self;
} /* -initWithStructure: */

-(TLDictionary *) methods
{
  return methods;
} /* -methods */

-(TLVector *) methodsNamed: (id <TLString>) n create: (BOOL) create
{
  TLVector *v = [methods objectForKey: n];

  if (!v && create)
    {
      v = [TLVector vector];
      [methods setObject: v forKey: n];
    }

  return v;
}

-(id <TLString>) outputDeclaration
{
  TLMutableString *s = nil;

  /* The output declaration can't be cached, because of the BT_RECV types
     and changing values of the OUTPUT_CURRENT_CONTEXT.  */
  if (vars_order)
    {
      id <TLEnumerator> e = [vars_order enumerator];
      OTMObjectVar *v;

      s = formac (s, @"{");
      while ((v = [e nextObject]))
	{
	  [@" " print: s quoted: NO];
	  [v outputDeclaration: s];
	  [@";" print: s quoted: NO];
	}

      /* Output the class information needed by the runtime.  */
      if ([structure meta] == ltt_class_state
	  && structure == [[structure meta] extensionNamed: nil])
	formac (s, @"%@", TRO_META_COMMON_INFORMATION);

      formac (s, @" }");
    }

  return s;
} /* -outputDeclaration */

-(id <TLString>) outputName
{
  return [structure outputName];
} /* -outputName */

-(id <TLString>) outputOffsetFrom: (LTTMeta *) m
{
  BOOL self_class_p = [[structure meta] classp];
  BOOL current_class_p = [m classp];
  id <TLString> tom_state_name
    = current_class_p ? tom_c_state_field_name : tom_i_state_field_name;
  id s = formac (nil, @"(struct %@ *) ", [structure outputExtensionStructName]);

  if (!current_class_p && self_class_p)
    {
      id <TLString> our_isa = formac (nil, @"%@->%@.%@", TO_NAME_SELF,
				      tom_state_name, TO_NAME_ISA);

      s = formac (s, @"((char *) %@ + %@->%@.%@->%@.info.%@->%@[%@])",
		  our_isa, our_isa, tom_c_state_field_name, TO_NAME_ISA,
		  tom_c_state_field_name, TO_NAME_EOT, TO_NAME_EOT_OFFSET,
		  [structure outputExtensionIdentityName]);
    }
  else
    s = formac (s, @"((char *) %@ + %@->%@.%@->%@.info.%@->%@[%@])",
		TO_NAME_SELF, TO_NAME_SELF, tom_state_name, TO_NAME_ISA,
		tom_c_state_field_name, TO_NAME_EOT, TO_NAME_EOT_OFFSET,
		[structure outputExtensionIdentityName]);
  return s;
} /* -outputOffsetFrom: */

-(void) startCompile
{
  formac (of, @"\n/* Start of %@.  */\n", ltt_ext_name (structure));
} /* -startCompile */

-(LTTExtension *) structure
{
  return structure;
} /* -structure */

-(OTMObjectVar *) variableNamed: (id <TLString>) n
{
  OTMObjectVar *v = vars_order ? [vars_keyed objectForKey: n] : nil;

  if (!v && vars_static)
    v = [vars_static objectForKey: n];
  if (!v && aliases)
    v = [aliases objectForKey: n];

  return v;
} /* -variableNamed: */

-(OTMObjectVar *) variableWithName: (id <TLString>) n
			      type: (id) t
			qualifiers: (otm_qualifiers) q
{
  OTMObjectVar *v;
  int i;

  /* XXX Handle qualifiers.  */

  if (![self cvDeclarationOK: n])
    return nil;

  i = [t minimumAlignment];
  if (i > vars_align)
    vars_align = i;

  v = [OTMObjectVar variableWithExtension: self name: n type: t
		    qualifiers: q];

  if (Q_STATIC (q))
    {
      if (!vars_static)
	vars_static = [TLDictionary dictionary];

      [vars_static setObject: v forKey: n];
    }
  else
    {
      if (!vars_order)
	{
	  vars_keyed = [TLDictionary dictionary];
	  vars_order = [TLVector vector];
	  [[[structure meta] semantics] setStatep];
	}

      [vars_keyed setObject: v forKey: n];
      [vars_order addElement: v];
    }

  return v;
}

@end
