/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: OTMTuple.m,v 1.28 1996/08/20 15:59:08 tiggr Exp $  */

#define OTMTUPLE_DECLARE_PRIVATE_METHODS
#import "OTMTuple.h"
#import "OTMDynamicType.h"
#import "OTMTypeTuple.h"
#import "OTMVariable.h"

@implementation OTMTuple

+(OTMTuple *) tuple
{
  return [[self gcAlloc] init];
} /* +tuple */

+(OTMTuple *) tupleWithSequence: (id) s
{
  return [[self gcAlloc] initWithSequence: s];
} /* +tupleWithSequence: */

-addElement: (id) o
{
  [elements addElement: o];
  return self;
} /* -addElement: */

-(BOOL) checkProperArgument: (OTMMethod *) m
{
  id <TLEnumerator> e = [elements enumerator];
  OTMVariable *v;
  BOOL r = NO;

  while ((v = [e nextObject]))
    r |= [v checkProperArgument: m];

  return r;
}

-(void) compileAddressListWithFirst: (id <TLString>) f
			  separator: (id <TLString>) s
{
  int i, n = [elements length];

  for (i = 0; i < n; i++)
    {
      [[elements _elementAtIndex: i] compileAddressListWithFirst: f
				     separator: s];
      f = s;
    }
}

-(void) compileAssignment: (OTMExpr *) rhs
{
  [rhs compileAssignmentToTuple: self];
} /* -compileAssignment: */

-(void) compileAssignmentToTuple: (OTMTuple *) lhs
{
  TLVector *vl = [(OTMTuple *) lhs elements];
  int i, n = [vl length];

  for (i = 0; i < n; i++)
    {
      OTMExpr *l = [vl _elementAtIndex: i];
      [l compileAssignment: implicit_cast ([elements _elementAtIndex: i],
					   [l type])];
    }
} /* -compileAssignmentToTuple: */

-(void) description: (id <TLMutableStream>) stream
{
  [super description: stream];
  formac (stream, @" %#", elements);
} /* -description: */

-(id) elaborate
{
  int i, n = [elements length];
  TLVector *ea = [TLVector vectorWithCapacity: n];
  OTMTuple *ta;

  for (i = 0; i < n; i++)
    {
      OTMExpr *ex = [[elements _elementAtIndex: i] elaborate];

      if ([ex isTuple] || [ex constantp])
	[ea addElement: ex];
      else
	[ea addElement:
	 emit_assignment (temp_something_with_type ([ex type], 1), ex)];
      }

  ta = [OTMTuple tupleWithSequence: ea];
  [ta setType: type];

  return ta;
} /* -elaborate */

-(TLVector *) elements
{
  return elements;
} /* -elements */

-(int) flatElementCount
{
  int c = 0, i, n = [elements length];

  for (i = 0; i < n; i++)
    c += [[elements _elementAtIndex: i] flatElementCount];

  return c;
}

-(void) gcReference
{
  MARK (elements);
  MARK (results);

  [super gcReference];
} /* -gcReference */

-(id) init
{
  [super init];

  elements = [TLVector vector];

  return self;
} /* -init */

-initWithSequence: (id) s
{
  if (![super init])
    return nil;

  elements = [TLVector vectorWithSequence: s];

  return self;
} /* -initWithSequence: */

-(BOOL) isTemporary
{
  return temporary;
} /* isTemporary */

-(BOOL) isTuple
{
  return YES;
} /* -isTuple */

-(OTMExpr *) lhsInvalid
{
  id <TLEnumerator> e = [elements enumerator];
  OTMExpr *o;

  while ((o = [e nextObject]))
    {
      id p = [o lhsInvalid];
      if (p)
	return p;
    }

  return nil;
} /* -lhsInvalid */

-(id <TLString>) name
{
  return @"tuple";
} /* -name */

-(id) outputDeclaration: (id) s
{
  return [self outputDeclaration: s separator: @","];
} /* -outputDeclaration: */

-(id) outputDeclaration: (id) s separator: (id <TLString>) sep
{
  int i, n = [elements length];

  for (i = 0; i < n; i++)
    {
      if (i)
	s = formac (s, @"%@\n\t", sep);
      s = [[elements _elementAtIndex: i] outputDeclaration: s];
    }
  return s;
} /* -outputDeclaration:separator: */

-(id) precompile
{
  int i, n = [elements length];

  for (i = 0; i < n; i++)
    {
      id p = [elements _elementAtIndex: i];
      id o = [p precompile];

      if (o != p)
	[elements _replaceElementAtIndex: i by: o];
    }
  return self;
} /* -precompile: */

-(TLCons *) resolveWithExpected: (TLCons *) expected
		    convertible: (OTMType *) to
			context: (OTMType *) cxt
			indices: (int *) indices
			  index: (int) index
{
  int  i, num_elts, our_index, is_dynamic;
  __typeof__ (indices) our_indices;
  TLVector *possible_elt_types;
  BOOL fully_typed, changes;

  if (type)
    return [super resolveWithExpected: expected convertible: to
		  context: cxt indices: indices index: index];

  our_index = index + 1;
  our_indices = alloca ((2 + index) * sizeof (*our_indices));
  memcpy (our_indices, indices, our_index * sizeof (*our_indices));

  num_elts = [elements length];
  possible_elt_types = [TLVector vectorWith: num_elts copies: nil];

  is_dynamic = (expected && ![expected cdr]
		&& [expected car] == the_dynamic_type);
  do
    {
      fully_typed = YES;
      changes = NO;

      /* Resolve our elements.  */
      for (i = 0; i < num_elts; i++)
	{
	  OTMExpr *o = [elements _elementAtIndex: i];
	  TLCons *pt;

	  if (!o)
	    {
	      if (!expected)
		[possible_elt_types _replaceElementAtIndex: i
				    by: CONS (the_any_type, nil)];
	      continue;
	    }

	  if (expected && !is_dynamic)
	    {
	      our_indices[our_index] = i;
	      pt = [o resolveWithExpected: expected convertible: to
		      context: cxt indices: our_indices index: our_index];
	    }
	  else
	    pt = [o resolveWithExpected: expected convertible: to
		    context: cxt indices: NULL index: -1];

	  if (!pt)
	    {
	      /* This element has no possible types.  Bummer.  */
	      return nil;
	    }

	  if (![o type])
	    fully_typed = NO;

	  if (expected && !is_dynamic)
	    {
	      if (i && !types_equal (pt, expected, expected))
		changes = YES;
	      expected = pt;
	      if (!is_dynamic)
		{
		  is_dynamic = (expected && ![expected cdr]
				&& [expected car] == the_dynamic_type);
		  if (is_dynamic)
		    changes = YES;
		}
	    }
	  else
	    {
	      OTMType *t = [o type];

	      /* Amend the list of possible types of this argument.  */
	      [possible_elt_types _replaceElementAtIndex: i
				  by: t ? CONS (t, nil) : pt];
	    }
	}
    } while (changes);

  if (fully_typed)
    {
      if (!expected || is_dynamic)
	type = [OTMTypeTuple typeTupleWithSequence:
			     [TLVector vectorByPerforming: @selector (car)
				     onElementsOfSequence: possible_elt_types]];
      else if ([expected cdr] || ![[expected car] isFullyDefinedType])
	{
	  ABORT ();
	}
      else
	type = [expected car];

      return [super resolveWithExpected: expected convertible: to
		    context: cxt indices: indices index: index];
    }

  error_for (self, @"unable to type tuple");

  return nil;
}

-(id <TLString>) result
{
  return [self result: 0];
} /* -result */

-(id <TLString>) result: (int) index
{
  return (results ? [results _elementAtIndex: index]
	  : [[elements _elementAtIndex: index] result]);
} /* -result: */

-(void) setIsTemporary
{
  temporary = !0;
} /* setIsTemporary */

@end
