<copyright> Runtime class.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1995, 1996 Pieter J. Schoenmakers.
    All rights reserved.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: Runtime.t,v 1.26 1996/09/05 15:00:09 tiggr Exp $</id>  </copyright>

<doc> The Runtime class provides an interface to the functionality in the
    runtime and other process related information.  </doc>
implementation class
Runtime: Conditions
{
  /******************** general ********************/

  <doc> The name of the host on which we're running.  </doc>
  static String hostname;

  <doc> The name under which this program was invoked.  </doc>
  public static String program_name;

  <doc> The long program name (i.e. C's argv[0]).  </doc>
  public static String long_program_name;

  <doc> The arguments as passed to {main}.  </doc>
  public static Array arguments;

  <doc> All the arguments, before any {load} method modified it.  </doc>
  public static Array all_arguments;

  /******************** signals ********************/

  // Maybe the information on signals should reside in a {Signal} class...
  <doc> Iff !0, quit ({SIGINT}, the user interrupt signal) is inhibited.
      </doc>
  private static int quit_inhibit;

  <doc> Iff !0, a {signal-int} will be raised when {quit_inhibit} and
      {panic_mode} again reach 0.  </doc>
  public static boolean quit_pending;

  <doc> Iff !0, any signal received (excluding the interrupt signal),
    other condition signaled or raised, or any object thrown will simply
    cause an abort.  This is used to protect critical sections in the
    runtime, such as during garbage collection or object allocation.  If
    {panic_mode} is set, {quit_inhibit} is implicitly set too.  </doc>
  private static int panic_mode;

  /******************** garbage collection statistics ********************/

  <doc> The number of objects allocated since the last partial garbage
      collection run.  In this respect, a partial run completing a full
      run is still considered a partial run.  </doc>
  public mutable static int gc_alloc_since_partial;

  <doc> The number of objects allocated since the previous completed run.
      This excludes the objects counted by {gc_alloc_since_partial}; it is
      adjusted after a run is initiated, before the run is actually
      started.  </doc>
  public mutable static int gc_alloc_since_total;

  <doc> Iff {TRUE} statistics on memory usage and the garbage collector
    will be emitted upon exit.  </doc>
  static boolean gc_stat_at_exit;

  <doc> Statistics on the garbage collector and allocator, in order: the
    number of gc runs; the number of runs which complete a full run; the
    number of object allocated; the number of objects deallocated; the
    (real, i.e. elapsed) time spent protecting, marking, and sweeping; and
    the time spent in all of gc (this is the sum of the previous three,
    plus overhead).  </doc>
  static int gc_num_runs, gc_num_complete;
  static long gc_num_alloc, gc_num_dealloc;
  static double gc_total_protect, gc_total_mark, gc_total_sweep, gc_total_all;

  /******************** garbage collection parameters ********************/

  <doc> Iff !0, garbage collection won't be run.  This is important
      during, for instance, enumerating a Container, since (most)
      enumerators can not handle the collection changing while they are
      enumerating.  </doc>
  public static int gc_inhibit;

  // Atomic garbage collection can be more efficient concerning memory
  // usage (since objects aren't gray by default, but white), but this has
  // not been added yet (even though it's trivial).
  // Tue May 21 15:55:32 1996, tiggr@tom.es.ele.tue.nl
  <doc> Iff {TRUE}, garbage collection will run atomically (with all its
      consequences), irrespective of the time constraint argument to
      {garbageCollect}.  The default value is {FALSE}.  </doc>
  public mutable static boolean gc_atomic;

  <doc> Iff {TRUE} all garbage will be cleaned upon exit.  This is a
    debugging tool mostly.  </doc>
  static boolean gc_full_at_exit;

  // Output to STDERR is wrong, but the result of the split of the
  // standard environment in two units (tom and too).  Maybe classes
  // needed by the runtime (time, streams, &c) should reside in tom too.
  //
  // The split has been undone, so this can be fixed.
  // Tue May 21 15:55:54 1996, tiggr@tom.es.ele.tue.nl
  <doc> The level of debugging garbage output by the garbage collector.
      Information is output to {stderr} stream provided by the C library.
      No information will be output if {gc_debug} is 0 or if the runtime
      was not compiled with the appropriate flags.  </doc>
  public mutable static int gc_debug;

  <doc> Threshold for {gc_alloc_since_partial} before a garbage collection
      run will be initiated.  If {gc_partial_threshold} is 0 (the default
      value), garbage collection is never run implicitly.  </doc>
  public mutable static int gc_partial_threshold;

  <doc> The time allowed for a partial garbage collection run when
      initiated by {gc_alloc_since_partial} exceeding
      {gc_partial_threshold}.  </doc>
  public mutable static double gc_partial_time_limit;

  <doc> When a partial garbage collection run is initiated and
      {gc_alloc_since_total} exceeds {gc_total_threshold}, the
      {gc_partial_time_limit} is ignored and instead the
      {gc_total_time_limit} is used.  If {gc_total_threshold} is 0 (the
      default value), it is ignored.  </doc>
  public mutable static int gc_total_threshold;

  <doc> The time limit used in case the condition described for
      {gc_total_threshold} applies.  </doc>
  public mutable static double gc_total_time_limit;
}

/********** entry and exit **********/

<doc> This method is invoked by the runtime.  Its main responsibility is
    to invoke the real main.  </doc>
int
      start (All, selector) (object, sel)
  arguments Array arguments
{
  [self exit [object perform sel with arguments]]
}

<doc> Normal level exit.  Cleaning up will be performed.  </doc>
void
  exit int rc
{
  [self willExit rc];
  [self fastExit rc]
}

<doc> Low level exit.  Usual functionality for cleaning up is avoided.  </doc>
extern void 
  fastExit int rc;

<doc> Perform all things necessary for a clean exit.  </doc>
void
  willExit int rc
{
  if (gc_full_at_exit)
    {
      if (!gc_atomic && (!!gc_alloc_since_partial || !!gc_alloc_since_total))
	[self garbageCollect];
      [self garbageCollect]
    };

  if (gc_stat_at_exit)
    {
      OutputStream s = [stdio err];

      [[[s print ("gc: runs: ", gc_num_runs)]
	print (" completed: ", gc_num_complete)] nl];
      [[[s print ("total: alloc: ", gc_num_alloc)]
	print (" dealloc: ", gc_num_dealloc)] nl];
      [[[[[s print ("time: ", gc_total_all)]
	  print (" protect: ", gc_total_protect)]
	 print (" mark: ", gc_total_mark)]
	print (" sweep: ", gc_total_sweep)] nl]
    }
}

<doc> Scan the arguments to the program for something telling us how to
    garbage collect.  </doc>
redefine void
  load MutableArray arguments
{
  int i, n = [arguments length];

  gc_total_threshold = 100000;

  for (i = 0; i < n;)
    {
      ByteString a = arguments[i];
      String s;

      if ([a length] > 0 && a[0] == ':')
	{
	  if ([":gc-debug" equal a])
	    {
	      gc_debug = 1;
	      [arguments removeElementAt i];
	      n--
	    }
	  else if ([":gc-pth" equal a] && i < n - 1)
	    {
	      gc_partial_threshold = [arguments[i + 1] unsignedIntValue];
	      [arguments removeElements (i, 2)];
	      n -= 2
	    }
	  else if ([":gc-ptl" equal a] && i < n - 1)
	    {
	      gc_partial_time_limit
		= 0.001 * double ([arguments[i + 1] unsignedIntValue]);
	      [arguments removeElements (i, 2)];
	      n -= 2
	    }
	  else if ([":gc-exit" equal a])
	    {
	      gc_full_at_exit = TRUE;
	      [arguments removeElementAt i];
	      n--
	    }
	  else if ([":gc-stat" equal a])
	    {
	      gc_stat_at_exit = TRUE;
	      [arguments removeElementAt i];
	      n--
	    }
	  else
	    i++
	}
      else
	i++
    }
}

<doc> Invoked by the runtime before the first load is invoked.  All it
    does is check for `:help'.  The reason this is not done in load is to
    be able to get some help before any negative side effects of load...
    </doc>
void
  preload Array arguments
{
  if ([arguments length] >= 1 && [":help" equal arguments[0]])
    {
      // Ugly!
      // Fri Aug 16 14:15:20 1996, tiggr@cobra.es.ele.tue.nl
      [BufferedStream load nil];

      OutputStream s = [BufferedStream with [stdio out]];
      Enumerator e = [[self classes] enumerator];
      class (State) c;
      boolean b;

      /* This is extremely expensive since, when dynamically resolved,
         this will build the dispatch tables of all class objects... */
      while ({(b, c) = [e next]; b})
	if ([c respondsTo selector ("r_help_r")])
	  // CCC This can not have an empty (void) context
	  // Fri Aug 16 14:01:02 1996, tiggr@cobra.es.ele.tue.nl
	  s = [c perform selector ("r_help_r") with s];

      [self exit 0]
    }
}

OutputStream
  help OutputStream s
{
  [[s print "tom.Runtime"] nl];
  [[s print "options: 
  :gc-debug	output gc debugging info to [stdio err]
  :gc-exit	perform a full gc run (possibly 2) upon exit
  :gc-pth <n>	set gc partial threshold
  :gc-ptl <n>	set gc partial run time limit to <n> milliseconds
  :gc-stat	output allocation and gc statistics upon exit"] nl];
  = s
}

/********** classes **********/
<doc> h{Classes}  </doc>

<doc> Return the dictionary of class objects.  </doc>
extern Mapped
  classes;

/********** selectors **********/
<doc> h{Selectors}  </doc>

<doc> Return the existing selector known by the {name}.  </doc>
extern selector
  selectorNamed String name;

<doc> Return the name of the {selector}.  </doc>
extern String
  nameOfSelector selector sel;

/********** garbage collection **********/
<doc> h{Garbage collection}  </doc>

<doc> Run the garbage collector to the end of a full garbage collection
    run.  </doc>
extern void
  garbageCollect;

<doc> Run the garbage collector for at most {time} seconds.  </doc>
extern void
  garbageCollect double time;

<doc> Increase the {gc_inhibit}.  This invocation should be matched by an
      invocation of {enableGC}.  </doc>
void
  disableGC
{
  gc_inhibit++
}

<doc> Decrease the {gc_inhibit}.  </doc>
void
  enableGC
{
  if (!gc_inhibit)
    [[Condition for self class program-condition
		message "attempt to decrease gc_inhibit below zero"] raise];
  gc_inhibit--
}

/********** process information **********/
<doc> h{Process information}  </doc>

<doc> Return the mutable dictionary holding the process environment.  </doc>
Keyed
  environment
{
  [self unimplemented cmd];

  = nil
}

<doc> Return the {hostname}.  If the class variable is not set, it is set
    once from gethostname(2).  </doc>
extern String
  hostname;

/********** signals **********/
<doc> h{Signal and condition handler}  </doc>

<doc> Accessor method for {quit_inhibit} which is private to the {Runtime}
    class to protect it against being mutated by subclasses but which can
    be freely read, hence this method.  </doc>
int
  quit_inhibit
{
  = quit_inhibit
}

<doc> Increase the {quit_inhibit} flag.  Any increase should be
    accompanied later on by the corresponding decrease.  </doc>
void
  quit_disable
{
  quit_inhibit++
}

<doc> Decrease the {quit_inhibit} flag, raising a postponsed {signal-int}
    if indicated by {quit_pending}.  </doc>
void
  quit_enable
{
  if (!quit_inhibit)
    [[Condition for self class lock-condition
		message "quit_enable while not disabled"] raise];

  if (!--quit_inhibit && quit_pending)
    {
      quit_pending = FALSE;
      [[Condition for self class signal-int
		  message "postponed interrupt"] raise]
    }
}

<doc> Accessor method for {panic_mode} which is private to the {Runtime}
    class to protect it against being mutated by subclasses but which can
    be freely read, hence this method.  </doc>
int
  panic_mode
{
  = panic_mode
}

<doc> Increase the {panic_mode} flag.  Any increase should be accompanied
    later on by the corresponding decrease.  </doc>
void
  panic_enable
{
  panic_mode++
}

<doc> Decrease the {panic_mode} flag, raising a {signal-int} if requested
    by {quit_pending}.  </doc>
void
  panic_disable
{
  if (!panic_mode)
    [[Condition for self class lock-condition
		message "panic_disable while not enabled"] raise];

  if (!--panic_mode && quit_pending && !quit_inhibit)
    {
      quit_pending = FALSE;
      [[Condition for self class signal-int
		  message "postponed interrupt"] raise]
    }
}

end;

implementation instance Runtime end;
