<copyright> Collection behavioural classes.
    Written by <a href="mailto:tiggr@es.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright &copy; 1995, 1996 Pieter J. Schoenmakers.
    All rights reserved.

    This file is part of TOM.  TOM is distributed under the terms of the
    TOM License, a copy of which can be found in the TOM distribution; see
    the file LICENSE.

    <id>$Id: collections.t,v 1.12 1996/09/02 18:31:06 tiggr Exp $</id>
    </copyright>

<doc> These collection definitions are based on the Smalltalk collections.
    </doc>

/******************** Collection ********************/

// Collection inherits from state to avoid conflicts with the write
// method, which otherwise need to be directed to Collection instead of
// State for every (conrete) Collection subclass.
// Mon Sep  2 20:04:10 1996, tiggr@tricky.es.ele.tue.nl
implementation class
Collection: State, Conditions, Enumerable

<doc> Return a new collection with the indicated elements.  The elements
    are boxed or unboxed as needed.  No other conversions are performed.
    </doc>
deferred instance (id)
  with dynamic elements;

end;

implementation instance
Collection

<doc> Make the receiving collection immutable.  This is irreversible.  It
    is a no-op for immutable collections.  </doc>
void
  freeze
{
  void
}

<doc> Return the number of elements in the {Collection}.  </doc>
deferred int
  length;

<doc> A {Collection} is not mutable.  </doc>
boolean
  mutable
{
  = FALSE
}

redefine OutputStream
  write OutputStream s
{
  Enumerator e = [self enumerator];
  boolean b;
  All o;

  s = [s print ('(', [[self kind] name])];

  while ({(b, o) = [e next]; b})
    s = [s print (' ', o)];

  = [s print ')']
}

end;

implementation class
MutableCollection: Collection

end;

implementation instance
MutableCollection

deferred void
  add All object;

void
  addElementsFrom Enumerable other
{
  [self addElementsFromEnumerator [other enumerator]]
}

void
  addElementsFromEnumerator Enumerator e
{
  Any object;

  while ({boolean b; (b, object) = [e next]; b})
    [self add object]
}

<doc> Force {freeze} to be undefined since each particular subclass must
    itself implement it.  </doc>
// CCC Actually, redeclaring something as deferred does not yet have the
// desired effect...  Wed Aug 7 14:28:22 1996, tiggr@cobra.es.ele.tue.nl
redefine deferred void
  freeze;

<doc> A {MutableCollection} is mutable.  </doc>
redefine boolean
  mutable
{
  = TRUE
}

end;

/******************** Keyed ********************/

<doc> A {Keyed} {Collection} stores elements on a key.  </doc>

implementation class
Keyed: Collection

end;

implementation instance
Keyed

deferred Any
  at All key;

end;

implementation class
MutableKeyed: Keyed, MutableCollection

end;

implementation instance
MutableKeyed

<doc> Add the {object}.  </doc>
deferred void
  add All object;

<doc> Remove the {object}.  </doc>
deferred void
  remove All object;

end;

/******************** Mapped ********************/

<doc> A {Mapped} {Collection} is a {Keyed} collection which stores (key,
    value) associations.  </doc>

implementation class
Mapped: Keyed

end;

implementation instance
Mapped

end;

implementation class
MutableMapped: Mapped, MutableKeyed

end;

implementation instance
MutableMapped

void
  add All value
{
  self[value] = value
}

deferred void
  set All value
   at All key;

end;

/******************** Ordered ********************/

<doc> A {Ordered} {Collection} maintains its elements in a specific order,
    though the time complexity or retrieving the nth object not
    necessarily independent of n.  </doc>

implementation class
Ordered: Collection

end;

implementation instance
Ordered

end;

implementation class
MutableOrdered: Ordered, MutableCollection

end;

implementation instance
MutableOrdered

end;

/******************** Indexed ********************/

<doc> An {Indexed} {Collection} maintains an association between integer
    indices and the objects it contains, with the promise that retrieving
    an object through the index is O(1) in time complexity.  </doc>

implementation class
Indexed: Ordered

end;

implementation instance
Indexed

<doc> Adjust the range ({start}, {len}) to fit the length of the receiving
    {Indexed} collection.  </doc>
(int, int)
  adjustRange (int, int) (start, len)
{
  int length = [self length];

  if (start >= length)
    {
      start = length;
      len = 0
    }
  else
    {
      if (len == -1)
	len = length - start;

      if (start < 0)
	{
	  len += start;
	  start = 0;

	  if (len < 0)
	    len = 0
	}
    };

  = (start, len)
}

<doc> Return the element at {index}.  If the receiving collection stores
    unboxed values, such as integers, the value returned is the element
    boxed. </doc>
deferred Any
  at int index;

<doc> Return the byte value of the element at {index}.  If the receiving
    collection stores objects, the {byteValue} of the element retrieved is
    actually returned.  </doc>
byte
  at int index
{
  // The object at INDEX is explicitly referred to as a Number here, since
  // the compiler must read the definition of Number to have seen the
  // `byteValue' method.
  Number n = self[index];

  = [n byteValue]
}

<doc> The following all follow the <basic type> {at} <index> stanza.  </doc>
char
  at int index
{
  = [self[index] charValue]
}

int
  at int index
{
  = [self[index] intValue]
}

long
  at int index
{
  = [self[index] longValue]
}

float
  at int index
{
  = [self[index] floatValue]
}

double
  at int index
{
  = [self[index] doubleValue]
}

<doc> Extract the elements from the receiving collection, as indicated by
    the return type.  The number of elements in the collection must match
    the number of expected elements.  </doc>
extern dynamic
  elements;

<doc> Like {elements}, but extract only the {num} elements starting at
    index {start}.  </doc>
extern dynamic
  elements (int, int) (start, num);

end;

implementation class
MutableIndexed: Indexed, MutableOrdered

end;

implementation instance
MutableIndexed

<doc> Store the {object} at {index} in the receiving collection.  </doc>
deferred void
  set All object
   at int index;

<doc> Swap the elements at the indices {i} and {j}.  </doc>
deferred void
  swap (int, int) (i, j);

<doc> Reverse the elements in the range starting at {start}, with length
    {len}.  </doc>
void
  reverse (int, int) (start, len)
{
  (start, len) = [self adjustRange (start, len)];
  int last = start + len - 1;

  while (start < last)
    {
      [self swap (start, last)];
      start++;
      last--
    }
}

end;
