/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: tr.m,v 1.47 1996/07/03 16:53:16 tiggr Exp $  */

#import "tr.h"
#import "tom/trt.h"

LTTInstance *tom_string_instance;

LTTUnit *app_unit, *top_unit;

TLSet *app_sel_arg_types;
TLSet *app_selectors;
TLSet *app_units;
TLSet *included_units;

TLSet *app_strings;

TLCons *super_refs;
int num_super_refs;

/* Flags.  */
int flag_no_def, flag_verbose_deferred;
int flag_instances_deferred, flag_instances_non_deferred;

TLCons *instances_deferred, *instances_non_deferred;

void
usage (int rc, char *s)
{
  if (s)
    formac (V_stderr_, @"%@: bad option/argument: %s\n", prog, s);

  formac (rc || s ? V_stderr_ : V_stdout_,
@"tr (%@), version %s\n\
usage: %@ [options] [unit]\n\
options:\n\
 --version	print version on stdout and exit\n\
 -I dir		add <dir> to the directories in which to find depended units\n\
 -u unit	name the unit to be resolved\n\

 -n		do not create definition file\n\
 -y unit	create def for dynamic loading into <unit>\n\
 -Y		create def for dynamic resolution\n\

 -v				increase verbosity\n\
 -v-deferred			report deferred selectors of each object\n\
 -v-deferred-instances		list deferred instances\n\
 -v-non-deferred-instances	list non-deferred instances\n\
", prog, long_version, prog);

  exit (rc);
}

void
tr_init (int argc, char **argv)
{
  lti_init (argc, argv);
}

void
tr_finish_init (void)
{
  lti_finish_init ();

  tom_string_instance = [ltt_unit_tom instanceNamed: TOM_CLASS_NAME_STRING];
  if (!tom_string_instance)
    error (@"Can't find instance %@.%@",
	   [ltt_unit_tom name], TOM_CLASS_NAME_STRING);
}

void
list_metas (TLCons *l)
{
  LTIMeta *m;

  while (l)
    {
      DECONS (l, m, l);
      formac (V_stderr_, @" %@.%@", [[[[m structure] unit] name] internal],
	      [[[m structure] name] internal]);
    }
}

int
main (int argc, char **argv)
{
  TLString *unit_name = nil, *app_unit_name = nil;
  id <TLEnumerator> e;
  int i, flag_Y = 0;
  LTTUnit *u;
  char *opt;

  tr_init (argc, argv);

  for (i = 1; i < argc; i++)
    if (argv[i][0] == '-')
      {
	BOOL ok = YES;

	switch (argv[i][1])
	  {
	  case '-':
	    if (!strcmp (argv[i], "--version"))
	      {
		if (flag_verbose)
		  printf ("tr, version %s\n", long_version);
		else
		  printf ("%s\n", short_version);
		exit (0);
	      }
	    break;

	  case 'h':
	  case '?':
	    /* Allow any option starting with a `h' or `?' to be an
	       indication of the need for help.  */
	    usage (0, NULL);
	    break;

	  case 'I':
	    opt = get_arg (&i, 2, argv, argc);
	    if (opt)
	      add_to_load_path ([[TLString stringWithCString: opt]
				 componentsSeparatedByString: @":"]);
	    break;

	  case 'n':
	    flag_no_def = 1;
	    break;

	  case 'u':
	    opt = get_arg (&i, 2, argv, argc);
	    if (opt)
	      unit_name = [TLString stringWithCString: opt];
	    break;

	  case 'v':
	    if (!strcmp (argv[i], "-v"))
	      flag_verbose++;
	    else if (!strcmp (argv[i], "-v-deferred"))
	      flag_verbose_deferred++;
	    else if (!strcmp (argv[i], "-v-deferred-instances"))
	      flag_instances_deferred++;
	    else if (!strcmp (argv[i], "-v-non-deferred-instances"))
	      flag_instances_non_deferred++;
	    else
	      ok = NO;
	    break;

	  case 'y':
	    if (!app_unit_name)
	      {
		opt = get_arg (&i, 2, argv, argc);
		if (opt)
		  app_unit_name = [TLString stringWithCString: opt];
	      }
	    else
	      ok = NO;
	    break;

	  case 'Y':
	    flag_Y = YES;
	    break;

	  default:
	    ok = NO;
	    break;
	  }

	if (!ok)
	  usage (1, argv[i]);
      }
    else if (unit_name)
      usage (1, argv[i]);
    else
      unit_name = [TLString stringWithCString: argv[i]];

  if (!NUM_ERRORS && !unit_name)
    {
      unit_name = basename_of_unit_file (@".");
      if (!unit_name)
	return 1;
    }

  tr_finish_init ();

  /* Iff needed, read the app unit.  */
  if (app_unit_name)
    {
      app_unit = [LTTUnit findUnitNamed: app_unit_name];
      if (!app_unit || NUM_ERRORS)
	return 1;
    }

  if (app_unit)
    {
      {
	e = [ltt_units valueEnumerator];
	while ((u = [e nextObject]))
	  if (u != ltt_builtin_unit && u != ltt_missing_unit)
	    lti_load_unit_info (u);
      }

      app_sel_arg_types = [TLSet setWithEnumerator:
				   [LTTSelArgTypes argumentTypes]];
      app_selectors = [TLSet setWithEnumerator: [LTISelector selectors]];
      app_units = [TLSet setWithEnumerator: [ltt_units valueEnumerator]];

      // XXX These can be removed...
      // Wed Jun 26 15:31:15 1996, tiggr@cobra.es.ele.tue.nl
      app_strings = [TLSet setWithEnumerator: [LTIStringCST allStrings]];

      /* Forget anything defined for the app unit which is not needed (or
         desired) for the top unit.  */
      [LTIStringCST sufferAmnesia];
      load_imps = nil;
    }

  /* Read the top unit.  */
  top_unit = [LTTUnit findUnitNamed: unit_name];
  if (!top_unit || NUM_ERRORS)
    return 1;
  if (app_unit && [app_unit dependsOn: top_unit])
    error (@"app unit `%@' depends on unit `%@'", app_unit_name, unit_name);

  if (flag_Y && !app_unit)
    app_unit = top_unit;

  {
    LTTUnit *u;

    e = [ltt_units valueEnumerator];
    included_units = [TLSet set];

    while ((u = [e nextObject]))
      if (u != ltt_builtin_unit && u != ltt_missing_unit
	  && (!app_units || ![app_units memq: u]))
	{
	  lti_load_unit_info (u);
	  [included_units addElement: u];
	}
  }

  if (NUM_ERRORS)
    return 1;

  {
    int num_included_selectors = 0, num_load_imps = 0;
    LTTSelector *main_sel = 0;
    LTTMeta *main_class = 0;
    TLString *decl_name, *def_name;
    TLFILEStream *decl, *def;
    time_t t = time (0);
    LTTSelArgTypes *args;
    LTIStringCST *cst;
    TLString *string;
    LTISelector *sel;
    LTTMeta *str;
    int i = 0;
    char *s;

    def_name = (id) formac (nil, @"%@/%@%@.c", [top_unit directory], unit_name,
			    TOM_RESOLVER_UNIT_SUFFIX);
    def = flag_no_def ? nil : open_output_file (def_name);

    decl_name = (id) formac (nil, @"%@/%@%@.h", [top_unit directory], unit_name,
			     TOM_RESOLVER_UNIT_SUFFIX);
    decl = open_output_file (decl_name);

    if (!decl || (!def && !flag_no_def))
      return 1;

    s = ctime (&t);
    s[24] = 0;
    string = formac (nil, @"/* Generated by tr (%@), on %s\n"
		     @"   version %s.  */\n", prog, s, long_version);
    if (def)
      {
	formac (def, @"%@#include %#\n", string, decl_name);
	if (!app_unit)
	  formac (def, @"#define _ trt_forward\n");
      }

    formac (decl, @"%@#ifndef __include_%@_resolved\n",
	    string, quote (unit_name));
    formac (decl, @"#define __include_%@_resolved\n", quote (unit_name));
    formac (decl, @"#include <tom/trt.h>\n", string);

    e = [LTTSelArgTypes argumentTypes];
    while ((args = [e nextObject]))
      {
	[args compileDeclaration: decl];
	if (def && (!app_sel_arg_types || ![app_sel_arg_types memq: args]))
	  [args compileDefinition: def];
      }

    e = [LTISelector selectors];
    while ((sel = [e nextObject]))
	{
	  [sel compileDeclaration: decl];
	  if (!app_selectors || ![app_selectors memq: sel])
	    {
	      num_included_selectors++;
	      if (def)
		[sel compileDefinition: def];
	    }
	}

    e = [LTTMeta metas];
    while ((str = [e nextObject]))
      [[str semantics] compileDeclaration: decl];

    if (def)
      {
	formac (def, @"static struct trtd_selectors selectors = {{%d, 0}, {\n",
		num_included_selectors);
	e = [LTISelector selectors];
	while ((sel = [e nextObject]))
	  if (!app_selectors || ![app_selectors memq: sel])
	    formac (def, @"&%@,\n", [[sel structure] outputDefinitionName]);
	formac (def, @"}};\n");

	[LTIMeta generateClassSequences];

	if (!app_unit)
	  {
	    formac (def,
		    @"struct trtd_selectors *trt_selectors = &selectors;\n");

	    formac (def, @"extern void _ ();\n");
	    e = [LTTMeta metas];
	    while ((str = [e nextObject]))
	      [[str semantics] compileBuckets: def];
	  }

	e = [LTTMeta metas];
	while ((str = [e nextObject]))
	  [[str semantics] compileTables: def];

	e = [LTTMeta metas];
	while ((str = [e nextObject]))
	  [[str semantics] compileDefinition: def];

	if (instances_deferred)
	  {
	    formac (V_stderr_, @"deferred instances:");
	    list_metas (instances_deferred);
	    formac (V_stderr_, @"\n");
	  }

	if (instances_non_deferred)
	  {
	    formac (V_stderr_, @"non-deferred instances:");
	    list_metas (instances_non_deferred);
	    formac (V_stderr_, @"\n");
	  }

	/* Output the string definitions.  */
	e = [LTIStringCST strings];
	while ((cst = [e nextObject]))
	  [cst compileDefinition: def];

	/* Output the string references.  */
	e = [LTIStringCST allStrings];
	while ((cst = [e nextObject]))
	  [cst compileReference: def];

	if (!app_unit)
	  {
	    /* Output the runtime TRT_METAS.  */
	    formac (def, @"\nstruct trt_metas trt_metas_def = {{%d, 0}, {",
		    [LTTMeta numMetas] / 2);
	    e = [LTTMeta metas];
	    while ((str = [e nextObject]))
	      if ([str classp])
		formac (def, @"\n(void *) &%@,", [str definitionName]);
	    formac (def, @"\n}};\n");

	    formac (def, @"struct trt_metas *trt_metas = &trt_metas_def;\n");
	  }

	if (flag_verbose)
	  {
	    int ns = [LTISelector numSelectors];
	    int nb = [mdt_buckets length];
	    int rns = ((ns + TRT_BUCKET_SIZE - 1) / TRT_BUCKET_SIZE
		       * TRT_BUCKET_SIZE);

	    formac (V_stdout_, @"%d classes, %d selectors, %d buckets, "
		    @"saving %2.1f%%/%2.1f%%\n", i, ns, nb,
		    100 - 100.0 * ((double) (nb * TRT_BUCKET_SIZE)
				   / (2 * i * ns)),
		    100 - 100.0 * ((double) (nb * TRT_BUCKET_SIZE
					     + 2 * i * rns / TRT_BUCKET_SIZE)
				   / (2 * i * ns)));
	  }

	/* Output information on the `load' implementations.  */
	while (load_imps)
	  {
	    LTIMethod *s;

	    DECONS (load_imps, s, load_imps);

	    if (!num_load_imps)
	      formac (def, @"\nstatic struct trtd_load_imp load_imps[] = {");
	    formac (def, @"%@{%@, (void *) &%@}",
		    num_load_imps ? @",\n" : @"\n",
		    [s name], [[[s meta] structure] referenceName]);
	    num_load_imps++;		
	  }
	if (num_load_imps)
	  formac (def, @"\n};\n");

	/* Output information on the super references.  */
	while (super_refs)
	  {
	    LTIMeta *subm, *superm;
	    TLCons *c;

	    DECONS (super_refs, c, super_refs);
	    DECONS (c, subm, superm);

	    if (!num_super_refs)
	      formac (def, @"\nstatic struct trtd_super super_refs[] = {");
	    formac (def, @"%@\n{(void *) &%@, (void *) &%@, (void *) &%@}",
		    num_super_refs ? @"," : @"",
		    [[subm structure] outputSuperReference: [superm structure]],
		    [[subm structure] metaReferenceName],
		    [[superm structure] metaReferenceName]);
	    num_super_refs++;
	  };
	if (num_super_refs)
	  formac (def, @"\n};\n");

	/* Deduce the main information, if we're supposed to have a `main'.  */
	if (!app_unit || app_unit == top_unit)
	  {
	    main_sel = [LTTSelector searchSelectorNamed: TO_SEL_MAIN];

	    if (!main_sel)
	      error (@"definition of main not found");
	    else
	      {
		TLVector *methods;

		methods = [TLVector vectorWithEnumerator:
				      [[main_sel semantics] methods]];
		if (![methods length])
		  error (@"definition of main not found");
		else if ([methods length] > 1)
		  {
		    /* This is an internal error as it does not check posing
		       and such.  */
		    internal (@"ambiguous definition of main");
		  }
		else
		  {
		    LTIMethod *m = [methods _elementAtIndex: 0];

		    main_class = [[m meta] structure];
		  }
	      }
	  }

	/* Output information on the units contained in this module.  */
	e = [included_units enumerator];
	while ((u = [e nextObject]))
	  {
	    int num_extensions = 0, num_metas = 0;
	    TLCons *included_extensions;
	    id <TLEnumerator> ie;

	    formac (def, @"\nstatic struct trtd_extension *"
		    @"trtd_extensions_%@[] = {\n", [[u name] external]);

	    ++search_mark;
	    included_extensions = [[ltt_class_all semantics]
				    compile: def unit: u extensionList:
				      [[ltt_instance_all semantics]
					compile: def unit: u extensionList: nil
					mark: search_mark]
				    mark: search_mark];
	    formac (def, @"};\n");

	    num_extensions = [included_extensions length];

	    ie = [u instances];
	    while ((str = [ie nextObject]))
	      {
		if (!num_metas)
		  formac (def, @"\nstatic struct trt_class *"
			  @"trtd_metas_%@[] = {", [[u name] external]);
		formac (def, @"%@(void *) &%@", num_metas ? @",\n" : @"\n",
			[str metaDefinitionName]);
		num_metas++;
	      }
	    if (num_metas)
	      formac (def, @"\n};\n");

	    formac (def, @"\nstatic struct trtd_unit %@ = {\n"
		    @"{%#, %d},\n%@, %d,\n%@, %d\n};\n",
		    [u outputDefinitionName], [[u name] internal],
		    [[[u name] internal] length],
		    (num_extensions
		     ? formac (nil, @"trtd_extensions_%@", [[u name] external])
		     : (id) @"0"), num_extensions,
		    (num_metas
		     ? formac (nil, @"trtd_metas_%@", [[u name] external])
		     : (id) @"0"), num_metas);
	  }

	i = 0;
	e = [[top_unit orderedUnits: nil] enumerator];
	formac (def, @"\nstatic struct trtd_unit *trtd_units[] = {");
	while ((u = [e nextObject]))
	  if (!included_units || [included_units memq: u])
	    {
	      formac (def, @"%@&%@", i ? @",\n" : @"\n",
		      [u outputDefinitionName]);
	      i++;
	    }
	formac (def, @"\n};\n");

	/* Output information on the dynamic selectors.  */
	if ([dynamic_selectors length])
	  {
	    LTISelector *alias;

	    formac (def, @"\nstatic struct trtd_selector_map "
		    @"dynamic_selectors[] = {");
	    e = [dynamic_selectors keyEnumerator];
	    while ((alias = [e nextObject]))
	      {
		LTISelector *orig = [dynamic_selectors objectForKey: alias];

		formac (def, @"\n{&%@, &%@},",
			[[alias structure] outputDefinitionName],
			[[orig structure] outputDefinitionName]);
	      }
	    formac (def, @"\n};\n");
	  }

	/* Output module description.  */
	formac (def, @"\nstruct trtd_module %s = {\ntrtd_units",
		(!app_unit || app_unit == top_unit ? "trt_main_module_info"
		 : TRT_MODULE_INFO));

	/* num_units */
	formac (def, @",\n%d", [included_units length]);

	/* selectors */
	formac (def, @",\n&selectors");

	/* dynamic_selectors, num_dynamic_selectors */
	if ([dynamic_selectors length])
	  formac (def, @",\ndynamic_selectors, %d",
		  [dynamic_selectors length]);
	else
	  formac (def, @",\n0, 0");

	/* super_refs, num_super_refs */
	if (num_super_refs)
	  formac (def, @",\nsuper_refs, %d", num_super_refs);
	else
	  formac (def, @",\n0, 0");

	/* load_imps, num_load_imps */
	if (num_load_imps)
	  formac (def, @",\nload_imps, %d", num_load_imps);
	else
	  formac (def, @",\n0, 0");

	/* main_selector, main_class */
	if (main_sel)
	  formac (def, @",\n&%@, (void *) &%@",
		  [main_sel outputDefinitionName], [main_class referenceName]);
	else
	  formac (def, @",\n0, 0");

	formac (def, @"\n};\n");

	[def close];
      }

    formac (decl, @"#endif\n");
    [decl close];

    if (NUM_ERRORS)
      {
	if (!flag_no_def && unlink ([def_name cString]))
	  error (@"unlink %@: %@", def_name, ERRMSG);
	if (unlink ([decl_name cString]))
	  error (@"unlink %@: %@", decl_name, ERRMSG);
      }
  }

  if (flag_verbose)
    formac (V_stderr_,
	    @"total %d line%s, %d error%s, %d warning%s\n",
	    total_num_lines, total_num_lines == 1 ? "" : "s",
	    total_num_errors, total_num_errors == 1 ? "" : "s",
	    total_num_warnings, total_num_warnings == 1 ? "" : "s");

  return !!NUM_ERRORS;
}
