/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: dload.c,v 1.23 1996/09/05 15:00:36 tiggr Exp $  */

#include "trt.h"
#include <tom/tom-r.h>
#include <stdio.h>  /* For ABORT.  */

static int dload_initialized = 0;

/* Iff !0, we're dynamically loading.  Otherwise, we're doing startup
   resolving.  */
static int dynamic_loading;

/* The number of existing buckets.  */
static int trt_num_buckets;

/* The number of buckets each dispatch table is supposed to have.  */
static int num_buckets = 0;

/* The method implementation hashtable.  This does not contain all methods.
   It is only used as a cache for metas_find_method().  */
struct hashtable method_table;

static struct trt_method_dispatch_table *empty_table;
static struct trt_method_dispatch_bucket *empty_bucket;
static struct trt_method_dispatch_table *install_table;
static struct trt_method_dispatch_bucket *install_bucket;

static void dtable_install (tom_object self, selector cmd, ...);

static __inline__ void
retain_bucket (struct trt_method_dispatch_bucket **b,
	       struct trt_method_dispatch_bucket *new)
{
  new->ref_count++;
  *b = new;
}

static __inline__ void
release_bucket (struct trt_method_dispatch_bucket **b)
{
  if (!*b || !(*b)->ref_count || (*b)->static_p)
    ABORT ();
  else
    {
      if (!--(*b)->ref_count)
	{
	  bzero (*b, sizeof (**b));
	  xfree (*b);
	}
      *b = 0;
    }
}

struct trt_method_dispatch_table *
dtable_resize (struct trt_method_dispatch_table *dt,
	       struct trt_method_dispatch_bucket *bucket)
{
  struct trt_method_dispatch_table *pdt = dt;
  int i, n = dt ? dt->st.num : 0;

  stalloc ((void *) &dt, sizeof (*dt), sizeof (dt->b[0]), num_buckets, 1);

  /* If the address of the table was changed, patch the owner of the
     buckets.  */
  if (pdt && pdt != dt)
    for (i = 0; i < n; i++)
      if (dt->b[i]->owner == pdt)
	dt->b[i]->owner = dt;

  /* Fill the extra slots created with the BUCKET.  */
  if (n != dt->st.num)
    for (i = n; i < num_buckets; i++)
      retain_bucket (&dt->b[i], bucket);

  return dt;
}

static void
dtable_free (struct trt_method_dispatch_table *dt)
{
  if (dt->st.cap)
    {
      int i;

      for (i = 0; i < dt->st.num; i++)
	if (!dt->b[i]->static_p)
	  release_bucket (&dt->b[i]);
      xfree (dt);
    }
}

static void
trt_include_selectors (struct trtd_selectors *sels)
{
  struct trt_method_dispatch_table *et, *it;
  int previous_num_buckets = num_buckets;
  int i, j;

  if (!dynamic_loading)
    {
      /* This is the easy case: we're invoked during startup to resolve the
         whole application.  This means we can directly use all the
         selectors.  */
      for (i = 1; i < sels->st.num; i++)
	sels->selectors[i]->sel_id = i;
      trt_selectors = sels;
      num_buckets = NUM_BUCKETS;
    }
  else if (sels->st.num)
    {
      /* Handle the (possibly new) selectors.  */
      for (i = 0; i < sels->st.num; i++)
	{
	  /* XXX Looping to find a match is not particularly fast...  */
	  for (j = 0; j < trt_selectors->st.num; j++)
	    if ((sels->selectors[i]->name.len
		 == trt_selectors->selectors[j]->name.len)
		&& !memcmp (sels->selectors[i]->name.s,
			    trt_selectors->selectors[j]->name.s,
			    sels->selectors[i]->name.len))
	      {
		sels->selectors[i]->sel_id
		  = trt_selectors->selectors[j]->sel_id;
		break;
	      }

	  if (j == trt_selectors->st.num)
	    {
	      sels->selectors[i]->sel_id = trt_selectors->st.num;
	      stalloc ((void *) &trt_selectors, sizeof (*trt_selectors),
		       sizeof (trt_selectors->selectors[0]),
		       trt_selectors->st.num + 1, 0);
	      trt_selectors->selectors[sels->selectors[i]->sel_id]
		= sels->selectors[i];
	    }
	}

      num_buckets = NUM_BUCKETS;
    }

  et = dtable_resize (empty_table, empty_bucket);
  it = dtable_resize (install_table, install_bucket);
  if (et != empty_table || it != install_table)
    {
      if (trt_metas)
	{
	  struct trt_class *m;

	  /* Patch all installed references to this table.  */
	  for (i = 0; i < trt_metas->st.num; i++)
	    for (j = 0, m = trt_metas->metas[i]; j < 2; j++, m = m->isa)
	      if (m->info.mdt == empty_table)
		m->info.mdt = et;
	      else if (m->info.mdt == install_table)
		m->info.mdt = it;
	      else if (num_buckets != previous_num_buckets)
		m->info.mdt = dtable_resize (m->info.mdt, empty_bucket);
	}
      empty_table = et;
      install_table = it;
    }
}

static void
trt_include_dynamic_selectors (struct trtd_selector_map *ds, int num_ds)
{
  int init_n = trt_dynamic_selectors ? trt_dynamic_selectors->st.num : 0;
  int i, j, n = init_n;

  for (i = 0; i < num_ds; i++)
    {
      for (j = 0; j < init_n; j++)
	if (trt_selectors_equal (trt_dynamic_selectors->maps[j]->actual,
				 ds[i].actual))
	  break;
      if (j == init_n)
	{
	  stalloc ((void *) &trt_dynamic_selectors,
		   sizeof (*trt_dynamic_selectors),
		   sizeof (trt_dynamic_selectors->maps[0]), n + 1, num_ds);
	  trt_dynamic_selectors->maps[n++] = &ds[i];
	}
    }
}

/* Given the method implementation IMP, return the definition of the method.
   This searches in from the meta M towards the supermost class.  */
static struct trtd_method *
metas_find_method (struct trt_class *m, void (*imp) (void), int k)
{
  int i, j;

  if (m->info.mark == k)
    return 0;
  m->info.mark = k;

  if (m->info.extensions)
    for (i = 0; i < m->info.extensions->st.num; i++)
      if (m->info.extensions->extensions[i]->methods)
	{
	  struct trtd_methods *ms = m->info.extensions->extensions[i]->methods;

	  for (j = 0; j < ms->num_methods; j++)
	    if (ms->methods[j].imp == imp)
	      return &ms->methods[j];
	}

  if (m->info.supers)
    for (i = 0; i < m->info.supers->st.num; i++)
      {
	struct trtd_method *found;

	found = metas_find_method (m->info.supers->metas[i], imp, k);
	if (found)
	  return found;
      }

  return 0;
}

static struct trtd_method *
find_method_info (struct trt_class *m, void (*imp) (void))
{
  struct trtd_method *mt = hash_find (&method_table, imp);

  if (!mt)
    {
      mt = metas_find_method (m, imp, ++trt_search_mark);
      if (!mt)
	/* XXX This could be a dtable_conflict which should be handled
           more elegantly...  */
	ABORT ();

      hash_add (&method_table, imp, mt);
    }

  return mt;
}

/* Add the meta M to the metas pointed to through PMETAS.  Return 0 if the
   addition was unnecessary, or 1 if it was added.  */
static int
metas_add (struct trt_metas **pmetas, struct trt_class *m)
{
  struct trt_metas *metas;
  int i, n = *pmetas ? (*pmetas)->st.num : 0;

  if (n)
    for (i = 0; i < n; i++)
      if ((*pmetas)->metas[i] == m)
	return 0;

  metas = stalloc ((void *) pmetas, sizeof (*metas), sizeof (metas->metas[0]),
		   n + 1, 4);

  metas->metas[n] = m;

  return 1;
}

static void
metas_install_default_mdt (struct trt_class *m, int k)
{
  if (m->info.mdt != install_table)
    {
      int i;

      if (m->info.mark == k)
	return;
      m->info.mark = k;

      if (m->info.mdt)
	dtable_free (m->info.mdt);
      m->info.mdt = install_table;

      /* Propagate to our subclasses.  */
      if (m->info.subs)
	for (i = 0; i < m->info.subs->st.num; i++)
	  metas_install_default_mdt (m->info.subs->metas[i], k);
    }
}

/* Mark the receiving meta M and its supers.  */
static void
meta_mark_supers (struct trt_class *m, int k)
{
  int i;

  if (m->info.mark == k)
    return;
  m->info.mark = k;

  if (m->info.supers)
    for (i = 0; i < m->info.supers->st.num; i++)
      meta_mark_supers (m->info.supers->metas[i], k);
}

/* Adjust the extension offset table and refvar offste table of the meta M
   to incorporate the state of the extension X.  */
static void
meta_add_extension_state (struct trt_class *m, struct trtd_extension *x)
{
  struct trt_extension_offset_table *eot;
  int i, n, x_offset, base;

  if (!x->num_vars
      || (m->info.eot && m->info.eot->st.num > *x->eid_in_a_global
	  && m->info.eot->offset[*x->eid_in_a_global]))
    ABORT ();

  if (TGC_CLASS_P (m->asi) && m->info.num_instances)
    fatal ("attempt to add state to class %s with %d allocated instances",
	   m->info.name.s, m->info.num_instances);

  /* Adjust the extension offset table.  */
  eot = stalloc ((void *) &m->info.eot, sizeof (*m->info.eot),
		 sizeof (eot->offset[0]), trt_num_eids, 1);

#if 0
  /* Adjust base if needed for one of the members of the extension (such as
     doubles or large pointers).
     XXX For now, only doubles are taken care of.  The rest will follow when
     I do the Alpha port.
     Tue Jul  2 13:54:49 1996, tiggr@cobra.es.ele.tue.nl  */
  base = m->info.instance_size;
  if (base & 7)
    {
      for (i = 0; i < x->num_vars; i++)
	if (x->vars[i].c.type == TRT_TE_DOUBLE
	    && ((base + x->vars[i].offset) & 7))
	  {
	    base = (base + 7) & ~7;
	    break;
	  }
    }
#else
  base = ((m->info.instance_size + x->state_align / 8 - 1)
	  & ~(x->state_align / 8 - 1));
  if (base < m->info.instance_size)
    ABORT ();
#endif

  eot->offset[*x->eid_in_a_global] = x_offset = base;
  m->info.instance_size += x->state_size;

  /* Adjust the reference variables offset table.  */
  for (i = 0, n = m->info.rvo ? m->info.rvo->st.num : 0; i < x->num_vars; i++)
    if (x->vars[i].c.type == TRT_TE_REFERENCE)
      {
	stalloc ((void *) &m->info.rvo, sizeof (*m->info.rvo),
		 sizeof (m->info.rvo->offset[0]), n + 1, 1);
	m->info.rvo->offset[n++] = x_offset + x->vars[i].offset;
      }
}

/* Invoke meta_add_extension_state to the meta M and all its subclasses.  */
static void
metas_add_extension_state (struct trt_class *m, struct trtd_extension *x, int k)
{
  int i;

  if (m->info.mark == k)
    return;
  m->info.mark = k;

  meta_add_extension_state (m, x);

  /* Propagate to our subclasses.  */
  if (m->info.subs)
    for (i = 0; i < m->info.subs->st.num; i++)
      metas_add_extension_state (m->info.subs->metas[i], x, k);
}

/* Add the state introduced by the class M and its superclasses to the newly
   added subclass TO.  */
static void
meta_inherit_state (struct trt_class *to, struct trt_class *m, int k)
{
  int i;

  if (m->info.mark == k)
    return;
  m->info.mark = k;

  /* Have our superclasses do this first (since State must end up as the
     first extension).  */
  if (m->info.supers)
    for (i = 0; i < m->info.supers->st.num; i++)
      meta_inherit_state (to, m->info.supers->metas[i], k);

  if (m->info.extensions)
    for (i = 0; i < m->info.extensions->st.num; i++)
      {
	struct trtd_extension *x = m->info.extensions->extensions[i];

	if (!x->state_size)
	  continue;

	/* State _must_ _always_ be the first.  */
	if (!to->info.instance_size
	    && !(x->eid_in_a_global == &_ei_i_tom_State
		 || x->eid_in_a_global == &_ei_c_tom_State))
	  ABORT ();

	metas_add_extension_state (to, x, ++trt_search_mark);
      }
}

/* Add any new superclasses introduced by the extension X to its meta.  */
static void
meta_add_extension_classes (struct trtd_extension *x)
{
  struct trt_class *m = x->meta;
  int i;

  /* Reshape the inheritance graph, inheriting state.  */
  for (i = 0; i < x->supers->st.num; i++)
    if (metas_add (&x->supers->metas[i]->info.subs, m))
      {
	/* Inherit the world from the newly added superclass.  */
	meta_mark_supers (m, ++trt_search_mark);
	meta_inherit_state (m, x->supers->metas[i], trt_search_mark);
	metas_add (&m->info.supers, x->supers->metas[i]);

	if (dynamic_loading)
	  {
	    /* Invalidate dispatch tables in ourselves and our subclasses.  */
	    metas_install_default_mdt (m, ++trt_search_mark);
	  }
      }

  if (x->state_size)
    {
      *x->eid_in_a_global = trt_num_eids++;
      metas_add_extension_state (m, x, ++trt_search_mark);
    }

  /* Reset the dispatch table.  */
  if (x->methods || !dynamic_loading)
    metas_install_default_mdt (m, ++trt_search_mark);

  /* Add the extension to its meta.  (Used by the dtable building.)  */
  i = m->info.extensions ? m->info.extensions->st.num : 0;
  stalloc ((void *) &m->info.extensions, sizeof (*m->info.extensions),
	   sizeof (m->info.extensions->extensions[0]), i + 1, 1);
  m->info.extensions->extensions[i] = x;
}

static struct trt_method_dispatch_table *
dtable_copy (struct trt_method_dispatch_table *ot)
{
  struct trt_method_dispatch_table *dt = 0;
  int i;

  stalloc ((void *) &dt, sizeof (*dt), sizeof (dt->b[0]), ot->st.num, 1);

  for (i = 0; i < dt->st.num; i++)
    retain_bucket (&dt->b[i], ot->b[i]);

  return dt;
}

static __inline__ void
dtable_put (struct trt_method_dispatch_table *dt,
	    int sel_id, void (*imp) (void))
{
  struct trt_method_dispatch_bucket *b;
  int bidx, midx;

  if (sel_id >= TRT_BUCKET_SIZE * dt->st.num)
    ABORT ();

  bidx = sel_id / TRT_BUCKET_SIZE;
  midx = sel_id % TRT_BUCKET_SIZE;

  b = dt->b[bidx];
  if (b->m[midx] != imp)
    {
      if (b->owner != dt)
	{
	  /* Copy on write.  */
	  struct trt_method_dispatch_bucket *new = xmalloc (sizeof (*new));

	  memcpy (&new->m[0], &b->m[0], sizeof (new->m));
	  new->static_p = new->ref_count = 0;
	  new->owner = dt;

	  release_bucket (&dt->b[bidx]);
	  retain_bucket (&dt->b[bidx], new);
	  b = dt->b[bidx];

	  trt_num_buckets++;
	}

      b->m[midx] = imp;
    }
}

static void
dtable_conflict (tom_object self, selector cmd)
{
  fatal ("conflict for selector %s of %s %s", cmd->name.s,
	 TGC_CLASS_P (self->isa->asi) ? "instance" : "class",
	 self->isa->info.name.s);
}

static void
dtable_deferred (tom_object self, selector cmd)
{
  fatal ("invocation of deferred method for selector %s of %s %s",
	 cmd->name.s, TGC_CLASS_P (self->isa->asi) ? "instance" : "class",
	 self->isa->info.name.s);
}

static void
metas_install_dtable (struct trt_class *m, int k)
{
  struct trt_class **inherited;
  int i, num_inherited = 0;

  if (m->info.mark == k || m->info.mdt != install_table)
    return;
  m->info.mark = k;

  /* First let our superclasses have a go at it.  */
  if (m->info.supers)
    for (i = 0; i < m->info.supers->st.num; i++)
      metas_install_dtable (m->info.supers->metas[i], k);
  else if (!m->info.extensions)
    {
      m->info.mdt = empty_table;
      return;
    }

  inherited = alloca (m->info.supers->st.num * sizeof (*inherited));
  for (i = 0; i < m->info.supers->st.num; i++)
    if (!(m->info.supers->metas[i]->info.mdt == empty_table
	  || m->info.supers->metas[i] == m))
      inherited[num_inherited++] = m->info.supers->metas[i];

  /* Build a dispatch table of the inherited functionality.  */
  if (num_inherited == 1)
    {
      /* This is an easy case: inheritance from a single superclass
         (a.k.a. single inheritance).  */
      m->info.mdt = dtable_copy (inherited[0]->info.mdt);
    }
  else if (num_inherited == 0)
    {
      /* No inheritance: start with an empty table.  XXX Actually, the copy
         should not be done since this case probably indicates an unused
         instance or a deferred instance (which should not be allocatable
         anyway) ...  */
      m->info.mdt = dtable_copy (empty_table);
    }
  else
    {
      /* General case: multiple inheritance.  */
      struct trt_method_dispatch_table *dt = 0;
      struct trt_method_dispatch_bucket *b;
      int super, bucket, method, inherited_super;

      m->info.mdt = stalloc ((void *) &dt, sizeof (*dt), sizeof (dt->b[0]),
			     empty_table->st.num, 1);

      for (bucket = 0; bucket < num_buckets; bucket++)
	{
	  /* First see if we can inherit any of the super buckets straight
             away without modification.  */
	  for (super = 0, b = 0; super < num_inherited; super++)
	    if (inherited[super]->info.mdt->b[bucket] != empty_bucket)
	      if (!b)
		b = inherited[super]->info.mdt->b[bucket];
	      else
		break;

	  if (super == num_inherited)
	    {
	      retain_bucket (&dt->b[bucket], b ? b : empty_bucket);
	      continue;
	    }

	  /* The fast way didn't work.  Do it the slow way.  */
	  inherited_super = -1;
	  for (method = 0; method < TRT_BUCKET_SIZE; method++)
	    {
	      void (*imp) (void) = 0;
	      int imp_super;

	      /* GGG NeXTstation NS3.3 with GNU CC 2.7.2
		 If this extra compound here is removed, the invocation of
		 dtable_put() afterwards is drawn into the loop!
		 Fri Jun 28 22:38:52 1996, tiggr@tricky.es.ele.tue.nl  */
	      {
		for (super = imp_super = 0; super < num_inherited; super++)
		  {
		    void (*this) (void);

		    this = inherited[super]->info.mdt->b[bucket]->m[method];

		    if (!this || this == (void *) dtable_install)
		      ABORT ();

		    if (this != (void *) trt_forward)
		      {
			if (!imp
			    || imp == (void *) dtable_deferred
			    /* XXX This is incorrect...  */
			    || imp == (void *) dtable_conflict)
			  {
			    imp_super = super;
			    imp = this;
			  }
			else if (imp != this
				 && this != (void *) dtable_deferred
				 /* XXX This is a hack!
				    v_load_r has 0 as sel_id  */
				 && !(this == (void *) dtable_conflict
				      && !bucket && !method))
			  {
			    /* This is inheritance of a different
                               implementation of the same method.  Let the
                               most specialized class (if they're related)
                               win.  XXX These semantics are questionable.  */
			    struct trtd_method *d_this, *d_imp;

			    d_this = find_method_info (inherited[super], this);
			    d_imp = find_method_info (m, imp);

			    if (trt_supermeta_star (d_this->meta, d_imp->meta,
						    ++trt_search_mark))
			      {
				imp_super = super;
				imp = this;
			      }
			    else if (!trt_supermeta_star (d_imp->meta,
							  d_this->meta,
							  ++trt_search_mark))
			      break;
			  }
		      }
		  }
	      }

	      imp = (super != num_inherited ? (void *) dtable_conflict
		     : imp ? imp : (void *) trt_forward);

	      if (imp != (void *) trt_forward)
		{
		  if (!dt->b[bucket])
		    {
		      if (inherited_super == -1)
			inherited_super = imp_super;
		      if (inherited[inherited_super]
			  ->info.mdt->b[bucket]->m[method] != imp)
			retain_bucket (&dt->b[bucket],
				       (inherited[inherited_super]
					->info.mdt->b[bucket]));
		    }

		  if (dt->b[bucket])
		    dtable_put (dt, method + bucket * TRT_BUCKET_SIZE, imp);
		}
	    }

	  if (!dt->b[bucket])
	    retain_bucket (&dt->b[bucket],
			   (inherited_super != -1
			    ? inherited[inherited_super]->info.mdt->b[bucket]
			    : empty_bucket));
	}
    }

  /* Override methods with our own definitions.  */
  if (m->info.extensions)
    for (i = 0; i < m->info.extensions->st.num; i++)
      {
	struct trtd_extension *x = m->info.extensions->extensions[i];
	int j;

	if (x->methods)
	  for (j = 0; j < x->methods->num_methods; j++)
	    dtable_put (m->info.mdt, x->methods->methods[j].sel->sel_id,
			x->methods->methods[j].imp ?: (void *) dtable_deferred);
      }

  /* Fix the dynamic selectors.  */
  if (trt_dynamic_selectors)
    {
      struct trtd_selector_map **maps = trt_dynamic_selectors->maps;

      for (i = 0; i < trt_dynamic_selectors->st.num; i++)
	{
	  void *act_imp = trt_lookup_super (m, maps[i]->actual);

	  if (act_imp == trt_forward || act_imp == dtable_deferred)
	    dtable_put (m->info.mdt, maps[i]->actual->sel_id,
			trt_lookup_super (m, maps[i]->dynamic));
	}
    }
}

static void
dtable_install (tom_object self, selector cmd, ...)
{
  void *args, *result;
  void_imp imp;

  metas_install_dtable (self->isa, ++trt_search_mark);

  imp = trt_lookup (self, cmd);
  args = __builtin_apply_args ();
  result = __builtin_apply (imp, args, trt_all_arguments_size (cmd));
  __builtin_return (result);
}

static void
trt_resolve_unit (struct trtd_unit *u)
{
  int i, n;

  /* Add new superclasses to each extension's meta.  */
  for (i = 0; i < u->num_extensions; i++)
    meta_add_extension_classes (u->extensions[i]);

  /* Remove the metas collection, as we won't be able to safely maintain
     it during resolution.  */
  if (trt_all_classes)
    trt_all_classes = 0;

  /* Add the classes from this unit to the recorded metas.  */
  n = trt_metas ? trt_metas->st.num : 0;
  stalloc ((void *) &trt_metas, sizeof (*trt_metas),
	   sizeof (trt_metas->metas[0]), n + u->num_classes, 0);
  for (i = 0; i < u->num_classes; i++)
    trt_metas->metas[n++] = u->classes[i];
}

void
trt_resolve_module (struct trtd_module *mi)
{
  int i;

  PANIC_MODE_P ();

  dynamic_loading = !!trt_metas;

  if (!empty_bucket)
    {
      int i;

      empty_bucket = xmalloc (sizeof (*empty_bucket));
      empty_bucket->static_p = empty_bucket->ref_count = 0;

      install_bucket = xmalloc (sizeof (*empty_bucket));
      install_bucket->static_p = install_bucket->ref_count = 0;

      trt_num_buckets += 2;

      for (i = 0; i < TRT_BUCKET_SIZE; i++)
	{
	  empty_bucket->m[i] = (void (*) ()) trt_forward;
	  install_bucket->m[i] = (void (*) ()) dtable_install;
	}

      empty_table = dtable_resize (0, empty_bucket);
      empty_bucket->owner = empty_table;
      install_table = dtable_resize (0, install_bucket);
      install_bucket->owner = install_table;

      method_table.equal = hash_equal_pointer;
      method_table.hash = hash_hash_pointer;
    }

  trt_include_selectors (mi->selectors);
  trt_include_dynamic_selectors (mi->dynamic_selectors,
				 mi->num_dynamic_selectors);

  for (i = 0; i < mi->num_units; i++)
    trt_resolve_unit (mi->units[i]);

  for (i = 0; i < mi->num_super_refs; i++)
    *mi->super_refs[i].ref = *mi->super_refs[i].super;

  if (!dynamic_loading)
    {
      /* These are missed since they have not a single extension.  */
      _md_m__builtin__Top.c_tom_State.info.mdt = install_table;
      _md_c__builtin__Top.c_tom_State.info.mdt = install_table;
    }

  PANIC_MODE_V ();
}

void
trt_execute_load_imps (struct trtd_module *mi, void *args)
{
  int i;

  for (i = 0; i < mi->num_load_imps; i++)
    mi->load_imps[i].imp (*mi->load_imps[i].class, SEL (v_load_r), args);
}

void *
c_tom_Bundle_p_load_r (tom_object self, selector cmd, tom_object object_name)
{
  void *handle;
  tom_int len;
  char *s;

  if (!dload_initialized)
    {
      dload_initialized = 1;
      dload_initialize ();
    }

  s = TRT_SEND ((pointer_imp), object_name, SEL (_pi__byteStringContents),
		&len);
  s = memcpy (alloca (len + 1), s, len);
  s[len] = 0;

  handle = dload_load (s);

  if (!handle)
    SIGNAL (self, c_tom_Conditions_error,
	    byte_string_with_c_string ("bundle load failed"));
  else
    {
      struct trtd_module *mi = dload_get_trt_module_info (handle);

      if (mi)
	{
	  trt_resolve_module (mi);
	  trt_execute_load_imps (mi, 0);
	}
    }

  return handle;
}
