/*
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: keyed.c,v 1.8 1996/08/22 16:37:32 tiggr Exp $  */

#include <string.h>
#include "trt.h"
#include <tom/tom-r.h>
#include "keyed.h"

bucket_elt *
bucket_elt_find (tom_object self, tom_object key)
{
  struct _es_i_tom_HashTable *ht = trt_ext_address (self, _ei_i_tom_HashTable);
  bucket_elt *be, **buckets = ht->buckets;
  tom_int hash;

  if (!ht->length)
    return 0;

  hash = TRT_SEND (, key, SEL (i_hash));
  hash = (unsigned) hash % ht->num_buckets;

  for (be = buckets[hash]; be; be = be->next)
    if (TRT_SEND ((byte_imp), key, SEL (o_equal_r), be->key))
      return be;

  return 0;
}

bucket_elt *
bucket_elt_find_eq (tom_object self, tom_object key)
{
  struct _es_i_tom_HashTable *ht = trt_ext_address (self, _ei_i_tom_HashTable);
  bucket_elt *be, **buckets = ht->buckets;
  tom_int hash;

  if (!ht->length)
    return 0;

  hash = HASH_POINTER (key);;
  hash = (unsigned) hash % ht->num_buckets;

  for (be = buckets[hash]; be; be = be->next)
    if (key == be->key)
      return be;

  return 0;
}

bucket_elt *
bucket_elt_remove (tom_object self, tom_object key)
{
  struct _es_i_tom_HashTable *ht = trt_ext_address (self, _ei_i_tom_HashTable);
  bucket_elt **pbe, **buckets = ht->buckets;
  tom_int hash;

  if (!ht->length)
    return 0;

  hash = TRT_SEND (, key, SEL (i_hash));
  hash = (unsigned) hash % ht->num_buckets;

  for (pbe = &buckets[hash]; *pbe; pbe = (void *) &(*pbe)->next)
    if (TRT_SEND ((byte_imp), key, SEL (o_equal_r), (*pbe)->key))
      {
	bucket_elt *be = *pbe;

	*pbe = be->next;
	be->next = NULL;
	ht->length--;

	return be;
      }

  return 0;
}

bucket_elt *
bucket_elt_remove_eq (tom_object self, tom_object key)
{
  struct _es_i_tom_HashTable *ht = trt_ext_address (self, _ei_i_tom_HashTable);
  bucket_elt **pbe, **buckets = ht->buckets;
  tom_int hash;

  if (!ht->length)
    return 0;

  hash = HASH_POINTER (key);
  hash = (unsigned) hash % ht->num_buckets;

  for (pbe = &buckets[hash]; *pbe; pbe = (void *) &(*pbe)->next)
    if (key == (*pbe)->key)
      {
	bucket_elt *be = *pbe;

	*pbe = be->next;
	be->next = NULL;
	ht->length--;

	return be;
      }

  return 0;
}

bucket_elt *
bucket_elt_find_create (tom_object self, tom_object key, int size)
{
  struct _es_i_tom_HashTable *ht = trt_ext_address (self, _ei_i_tom_HashTable);
  tom_int hash = TRT_SEND (, key, SEL (i_hash));
  bucket_elt *be, **buckets;

  if (!ht->num_buckets || ht->length / ht->num_buckets)
    TRT_SEND (, self, SEL (v_resize));
  buckets = ht->buckets;

  hash = (unsigned) hash % ht->num_buckets;
  for (be = buckets[hash]; be; be = be->next)
    if (TRT_SEND ((byte_imp), key, SEL (o_equal_r), be->key))
      return be;

  be = xmalloc (size);
  if (size != sizeof (*be))
    BZERO (be + 1, size - sizeof (*be));
  be->key = key;
  be->next = buckets[hash];
  buckets[hash] = be;
  ht->length++;

  return be;
}

bucket_elt *
bucket_elt_find_create_eq (tom_object self, tom_object key, int size)
{
  struct _es_i_tom_HashTable *ht = trt_ext_address (self, _ei_i_tom_HashTable);
  tom_int hash = HASH_POINTER (key);
  bucket_elt *be, **buckets;

  if (!ht->num_buckets || ht->length / ht->num_buckets)
    TRT_SEND (, self, SEL (v_resize));
  buckets = ht->buckets;

  hash = (unsigned) hash % ht->num_buckets;
  for (be = buckets[hash]; be; be = be->next)
    if (key == be->key)
      return be;

  be = xmalloc (size);
  if (size != sizeof (*be))
    BZERO (be + 1, size - sizeof (*be));
  be->key = key;
  be->next = buckets[hash];
  buckets[hash] = be;
  ht->length++;

  return be;
}

void
i_tom_HashTable_v_dealloc (tom_object self, selector cmd)
{
  struct _es_i_tom_HashTable *ht = trt_ext_address (self, _ei_i_tom_HashTable);
  set_elt **buckets = ht->buckets;
  tom_int i, n = ht->num_buckets;

  for (i = 0; i < n; i++)
    while (buckets[i])
      {
	set_elt *be = buckets[i];

	buckets[i] = be->c.next;
	xfree (be);
      }

  xfree (ht->buckets);
}

void
i_tom_HashTable_v_gc_mark_elements (tom_object self, selector cmd)
{
  struct _es_i_tom_HashTable *ht = trt_ext_address (self, _ei_i_tom_HashTable);
  set_elt *be, **buckets = ht->buckets;
  tom_int i, n = ht->num_buckets;

  for (i = 0; i < n; i++)
    for (be = buckets[i]; be; be = be->c.next)
      if (TGC_COLOUR (be->c.key->asi) == TGC_WHITE)
	trt_make_gray (be->c.key);
}

/* XXX Maybe panic mode should be ON during this rather atomic resizing
   operation.  On the other hand, maybe only quit needs to be inhibited.
   Or maybe a SIGINT should be signaled, and a sensible program actually
   doing something with a quit should just set the quit request flag.  */
void
i_tom_HashTable_v_resize (tom_object self, selector cmd)
{
  struct _es_i_tom_HashTable *ht = trt_ext_address (self, _ei_i_tom_HashTable);
  bucket_elt *be, *nbe, **buckets, **previous = ht->buckets;
  tom_int i, previous_num = ht->num_buckets;

  ht->num_buckets = previous_num ? previous_num * 4 : 4;
  buckets = ht->buckets = xcalloc (ht->num_buckets, sizeof (*previous));

  for (i = 0; i < previous_num; i++)
    for (be = previous[i]; be; be = nbe)
      {
	tom_int hash = TRT_SEND (, be->key, SEL (i_hash));

	hash = (unsigned) hash % ht->num_buckets;
	nbe = be->next;
	be->next = buckets[hash];
	buckets[hash] = be;
      }

  xfree (previous);
}

void
i_tom_EqHashTable_v_resize (tom_object self, selector cmd)
{
  struct _es_i_tom_HashTable *ht = trt_ext_address (self, _ei_i_tom_HashTable);
  bucket_elt *be, *nbe, **buckets, **previous = ht->buckets;
  tom_int i, previous_num = ht->num_buckets;

  ht->num_buckets = previous_num ? previous_num * 4 : 4;
  buckets = ht->buckets = xcalloc (ht->num_buckets, sizeof (*previous));

  for (i = 0; i < previous_num; i++)
    for (be = previous[i]; be; be = nbe)
      {
	tom_int hash = HASH_POINTER (be->key);

	hash = (unsigned) hash % ht->num_buckets;
	nbe = be->next;
	be->next = buckets[hash];
	buckets[hash] = be;
      }

  xfree (previous);
}

void
tgc_mark_containers (void)
{
  struct _es_i_tom_HashTable *ht
    = trt_ext_address (c_tom_Container_all_containers, _ei_i_tom_HashTable);
  bucket_elt **pbe, **buckets = ht->buckets;
  tom_int i, n = ht->num_buckets;

  for (i = 0; i < n; i++)
    for (pbe = &buckets[i]; *pbe;)
      {
	tom_object o = (*pbe)->key;
	bucket_elt *be = *pbe;

	if (TGC_COLOUR (o->asi) == TGC_WHITE)
	  {
	    o->asi &= ~TGC_ASI_CONTAINER_P;
	    *pbe = be->next;
	    if (!--ht->length)
	      break;
	  }
	else
	  {
	    TRT_SEND ((void_imp), o, SEL (v_gc_container_mark_elements));
	    pbe = (void *) &(*pbe)->next;
	  }
      }
}

tom_byte
i_tom_HashTable__orip__nextKey__ip_ (tom_object self, selector cmd,
				     int bucket, struct bucket_elt *be,
				     tom_object *object, tom_int *nbucket,
				     struct bucket_elt **nbe)
{
  struct _es_i_tom_HashTable *ht = trt_ext_address (self, _ei_i_tom_HashTable);
  struct bucket_elt **buckets = ht->buckets;
  tom_byte r = 0;

  if (bucket == ht->num_buckets)
    {
      *object = 0;
      be = 0;
    }
  else
    {
      if (be)
	{
	  be = be->next;
	  if (!be)
	    bucket++;
	}

      if (!be)
	{
	  while (bucket < ht->num_buckets && !buckets[bucket])
	    bucket++;

	  if (bucket < ht->num_buckets)
	    be = buckets[bucket];
	  else
	    *object = 0;
	}

      if (be)
	{
	  *object = be->key;
	  r = 1;
	}
    }


  *nbucket = bucket;
  *nbe = be;
  return r;
}
