/* C routines for streams and date.
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>
          and Michael L.H. Brouwer <michael@thi.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: timespace.c,v 1.13 1996/08/20 15:59:38 tiggr Exp $  */

#include "trt.h"
#include <tom/tom-r.h>
#include <config/config.h>

#if HAVE_LIBC_H
#include <libc.h>
#endif

#if HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <stdio.h>
#include <sys/file.h>
#include <sys/time.h>

#define UNIX_OFFSET	-978307200.0

double
c_tom_Date_d_timeIntervalSinceReferenceDate (tom_object self, selector *_cmd)
{
  struct timeval tv;

  gettimeofday (&tv, NULL);
  return UNIX_OFFSET + tv.tv_sec + tv.tv_usec * 1e-6;
}

double
c_tom_Date_d_relativeTimeIntervalSinceNow (tom_object self, selector *_cmd)
{
  struct _es_c_tom_Date *this;
  struct timeval tv;

  this = trt_ext_address (self, _ei_c_tom_Date);

  gettimeofday (&tv, NULL);
  return (UNIX_OFFSET - this->relative_offset) + tv.tv_sec + tv.tv_usec * 1e-6;
}

tom_object
i_tom_File_r_reopen (tom_object self, selector cmd)
{
  struct _es_i_tom_File *fs = trt_ext_address (self, _ei_i_tom_File);
  struct _es_i_tom_Descriptor *bs;
  char input_p, output_p, exists;
  tom_int l;
  char *s;

  bs = trt_ext_address (self, _ei_i_tom_Descriptor);

  if (bs->descriptor != -1)
    TRT_SEND (, self, SEL (v_close));

  if (bs->descriptor != -1)
    RAISE (self, c_tom_Conditions_file_error,
	   byte_string_with_c_string ("unable to close for reopen"));

  input_p = !!(fs->flags & c_tom_File_OPEN_INPUT);
  output_p = !!(fs->flags & c_tom_File_OPEN_OUTPUT);

  /* Get the filename in a way C can understand.  */
  s = TRT_SEND ((pointer_imp), fs->name, SEL (_pi__byteStringContents), &l);
  s = memcpy (alloca (1 + l), s, l);
  s[l] = 0;

  if (!access (s, F_OK))
    {
      /* It exists.  */
      if (fs->flags & c_tom_FileUser_EXIST_NOTHING)
	return 0;
      else if (fs->flags & c_tom_FileUser_EXIST_RAISE)
	RAISE (self, c_tom_Conditions_file_error,
	       byte_string_with_c_string ("file exists"));
      exists = 1;
    }
  else
    {
      /* It does not exist (as far as we're concerned).  */
      if (fs->flags & c_tom_FileUser_NOT_EXIST_NOTHING)
	return 0;
      else if (fs->flags & c_tom_FileUser_NOT_EXIST_RAISE)
	RAISE (self, c_tom_Conditions_file_error,
	       byte_string_with_c_string (OS_ERROR_MSG));
      exists = 0;
    }


  if (input_p)
    {
      if (output_p)
	{
	  /* Read and write.  */
	  ABORT ();
	}
      else
	{
	  /* Read only.  */
	  if (!exists)
	    return 0;

	  bs->descriptor = open (s, O_RDONLY, 0);
	  if (bs->descriptor == -1)
	    RAISE (self, c_tom_Conditions_file_error,
		   byte_string_with_c_string (OS_ERROR_MSG));
	}
    }
  else if (output_p)
    {
      int flags = O_WRONLY | O_TRUNC;

      if (fs->flags & c_tom_FileUser_NOT_EXIST_CREATE)
	flags |= O_CREAT;
      if (fs->flags & c_tom_FileUser_APPEND)
	flags |= O_APPEND;

      /* XXX The 0666 should be an argument to this method.  */
      bs->descriptor = open (s, flags, 0666);
      if (bs->descriptor == -1)
	RAISE (self, c_tom_Conditions_file_error,
	       byte_string_with_c_string (OS_ERROR_MSG));
    }
  else
    {
      /* Probe only, which has been done already.  */
    }

  return self;
}

tom_int
i_tom_ByteStream_i_writeRange__ii__from_r
  (tom_object self, selector cmd,
   tom_int start, tom_int length, tom_object buf)
{
  struct _es_i_tom_Descriptor *this;
  char *s;

  this = trt_ext_address (self, _ei_i_tom_Descriptor);
  s = TRT_SEND ((pointer_imp), buf, SEL (_pi__pointerToElements__ii_),
		start, length, &length);

  return length <= 0 ? 0 : write (this->descriptor, s + start, length);
}

tom_int
i_tom_MutableByteArray_i_writeRange__ii__from_r
  (tom_object self, selector cmd, tom_int start, tom_int length, tom_object buf)
{
  tom_int n;
  tom_byte *s;
  struct _es_i_tom_Array *this;
  struct _es_i_tom_MutableArray *that;

  this = trt_ext_address (self, _ei_i_tom_Array);
  that = trt_ext_address (self, _ei_i_tom_MutableArray);

  s = TRT_SEND ((pointer_imp), buf, SEL (_pi__byteStringContents), &n);

  MODIFY_RANGE (start, length, n);

  if (length)
    {
      if (this->length + length >= that->capacity)
	{
	  tom_int new_cap = this->length + length + 80;

	  this->contents = xrealloc (this->contents, new_cap);
	  that->capacity = new_cap;
	}

      memcpy ((tom_byte *) this->contents + this->length, s + start, length);
      this->length += length;
    }

  return length;
}

tom_byte
i_tom_ByteStream_b_read (tom_object self, selector cmd)
{
  struct _es_i_tom_Descriptor *this;
  tom_byte c;
  int i;

  this = trt_ext_address (self, _ei_i_tom_Descriptor);
  i = read (this->descriptor, &c, 1);
  if (i != 1)
    if (i == 0)
      RAISE (self, c_tom_Conditions_stream_eos, 0);
    else if (i < 0)
      RAISE (self, c_tom_Conditions_stream_error,
	     byte_string_with_c_string (OS_ERROR_MSG));

  return c;
}

tom_int
i_tom_ByteStream_i_read (tom_object self, selector cmd)
{
  struct _es_i_tom_Descriptor *this;
  tom_byte c;
  int i;

  this = trt_ext_address (self, _ei_i_tom_Descriptor);
  i = read (this->descriptor, &c, 1);
  if (i < 0)
    RAISE (self, c_tom_Conditions_stream_eos,
	   byte_string_with_c_string (OS_ERROR_MSG));

  return i ? c : -1;
}

void
i_tom_ByteStream_v_write_b (tom_object self, selector cmd, tom_byte b)
{
  struct _es_i_tom_Descriptor *this;
  int i;

  this = trt_ext_address (self, _ei_i_tom_Descriptor);
  i = write (this->descriptor, &b, 1);
  if (i != 1)
    if (i == 0)
      RAISE (self, c_tom_Conditions_stream_eos, 0);
    else
      RAISE (self, c_tom_Conditions_stream_error,
	     byte_string_with_c_string (OS_ERROR_MSG));
}

tom_int
i_tom_ByteStream_i_write_b (tom_object self, selector cmd, tom_byte b)
{
  struct _es_i_tom_Descriptor *this;
  int i;

  this = trt_ext_address (self, _ei_i_tom_Descriptor);
  i = write (this->descriptor, &b, 1);
  if (i < 0)
    RAISE (self, c_tom_Conditions_stream_error,
	   byte_string_with_c_string (OS_ERROR_MSG));

  return i;
}

tom_int
i_tom_MutableByteArray_i_readRange__ii__fromByteStream_r
  (tom_object self, selector cmd, tom_int start, tom_int num, tom_object file)
{
  struct _es_i_tom_MutableArray *m;
  struct _es_i_tom_Array *a;
  int i, fd = TRT_SEND ((int_imp), file, SEL (i_descriptor));

  m = trt_ext_address (self, _ei_i_tom_MutableArray);
  a = trt_ext_address (self, _ei_i_tom_Array);

  if (start < 0 || num < 0 || start > a->length)
    RAISE (self, c_tom_Conditions_program_condition,
	   byte_string_with_c_string ("bad arguments"));

  if (m->capacity < start + num)
    trt_array_resize (a, m, 1, start + num - a->length);

  i = read (fd, (tom_byte *) a->contents + start, num);
  if (i < 0)
    SIGNAL (file, c_tom_Conditions_stream_error,
	    byte_string_with_c_string (OS_ERROR_MSG));

  if (a->length < start + i)
    a->length = start + i;

  return i;
}

void
init_streams (void)
{
  struct _es_i_tom_Descriptor *this;

  c_tom_stdio_in = (void *) TRT_SEND (, _mr_c_tom_ByteStream, SEL (r_alloc));
  this = trt_ext_address (c_tom_stdio_in, _ei_i_tom_Descriptor);
  this->descriptor = 0;

  c_tom_stdio_out = (void *) TRT_SEND (, _mr_c_tom_ByteStream, SEL (r_alloc));
  this = trt_ext_address (c_tom_stdio_out, _ei_i_tom_Descriptor);
  this->descriptor = 1;

  c_tom_stdio_err = (void *) TRT_SEND (, _mr_c_tom_ByteStream, SEL (r_alloc));
  this = trt_ext_address (c_tom_stdio_err, _ei_i_tom_Descriptor);
  this->descriptor = 2;
}

void
tini_streams (void)
{
  if (c_tom_stdio_in)
    TRT_SEND (, c_tom_stdio_in, SEL (v_close));
  if (c_tom_stdio_out)
    TRT_SEND (, c_tom_stdio_out, SEL (v_close));
  if (c_tom_stdio_err)
    TRT_SEND (, c_tom_stdio_err, SEL (v_close));
}

void
c_tom_stdio_v_close_i (tom_object self, selector cmd, tom_int d)
{
  if (close (d))
    SIGNAL (self, c_tom_Conditions_stream_error,
	    byte_string_with_c_string (OS_ERROR_MSG));
}

i_tom_OutputStream
i_tom_OutputStream_r_print_p (tom_object self, selector cmd, void *p)
{
  POINTER_AS_INT np = (POINTER_AS_INT) p;
  struct _es_i_tom_Array *this;
  i_tom_ByteString str;
  int i;

  /* XXX Maybe printing should not go through a string, as that might be
     very expensive.  The message overhead is considerable, especially if
     most numbers are small (i.e. less than a few digits).  */
  str = new_byte_string (&this, 2 * sizeof (void *));

  for (i = 2 * sizeof (void *) - 1; i >= 0; i--, np >>= 4)
    ((char *) this->contents)[i] = "0123456789abcdef"[np & 0xf];

  return (void *) TRT_SEND (, str, SEL (r_write_r), self);
}

#define PRINT(T, FT)  \
i_tom_OutputStream    							\
C2 (i_tom_OutputStream_r_print_, FT)					\
  (tom_object self, selector cmd, T arg)				\
{		      							\
  struct _es_i_tom_Array *this;						\
  int n = 28;								\
  i_tom_ByteString str = new_byte_string (&this, n);			\
  int i = n - 1, neg = 0;						\
		      							\
  if (arg < 0)	      							\
    arg = -arg, neg = 1;						\
		      							\
  do		      							\
    {		      							\
      ((char *) this->contents)[i] = (arg % 10) + '0';			\
      arg /= 10;      							\
      i--;	      							\
    } while (arg);    							\
		      							\
  if (neg)	      							\
    ((char *) this->contents)[i--] = '-';				\
		      							\
  this->length = n - i - 1;						\
  memmove (this->contents, (char *) this->contents + i + 1, this->length);  \
									\
  return (void *) TRT_SEND (, str, SEL (r_write_r), self);		\
}

PRINT (tom_byte, o)
PRINT (tom_char, c)
PRINT (tom_int, i)
PRINT (tom_long, l)

static i_tom_OutputStream
print_fp (tom_object self, double arg)
{
  struct _es_i_tom_Array *this;
  int n = 60;
  i_tom_ByteString str = new_byte_string (&this, n);

  sprintf (this->contents, "%f", arg);
  this->length = strlen (this->contents);

  return (void *) TRT_SEND (, str, SEL (r_write_r), self);
}

i_tom_OutputStream
i_tom_OutputStream_r_print_d (tom_object self, selector cmd, double arg)
{
  return print_fp (self, arg);
}

i_tom_OutputStream
i_tom_OutputStream_r_print_f (tom_object self, selector cmd, float arg)
{
  return print_fp (self, arg);
}

tom_object
i_tom_Sink_r_print_x (tom_object self, selector cmd, ...)
{
  /* Nothing much to do for a sink...  */
  return self;
}

i_tom_OutputStream
i_tom_OutputStream_r_print_x (tom_object self, selector cmd, ...)
{
  va_list ap;
  int i;

  va_start (ap, cmd);

  for (i = 0; i < cmd->in->num; i++)
    switch (cmd->in->args[i])
      {
      case TRT_TE_BOOLEAN:
	self = TRT_SEND ((reference_imp), self, SEL (r_print_o),
			 va_arg (ap, tom_byte));
	break;

      case TRT_TE_BYTE:
	self = TRT_SEND ((reference_imp), self, SEL (r_print_b),
			 va_arg (ap, tom_byte));
	break;

      case TRT_TE_CHAR:
	self = TRT_SEND ((reference_imp), self, SEL (r_print_c),
			 va_arg (ap, tom_char));
	break;

      case TRT_TE_INT:
	self = TRT_SEND ((reference_imp), self, SEL (r_print_i),
			 va_arg (ap, tom_int));
	break;

      case TRT_TE_LONG:
	self = TRT_SEND ((reference_imp), self, SEL (r_print_l),
			 va_arg (ap, tom_long));
	break;

      case TRT_TE_FLOAT:
	self = TRT_SEND ((tom_object (*) (tom_object, selector, tom_float)),
			 self, SEL (r_print_f), va_arg (ap, tom_float));
	break;

      case TRT_TE_DOUBLE:
	self = TRT_SEND ((reference_imp), self, SEL (r_print_d),
			 va_arg (ap, tom_double));
	break;

      case TRT_TE_POINTER:
	self = TRT_SEND ((reference_imp), self, SEL (r_print_p),
			 va_arg (ap, void *));
	break;

      case TRT_TE_REFERENCE:
	self = TRT_SEND ((reference_imp), self, SEL (r_print_r),
			 va_arg (ap, tom_object));
	break;

      default:
	fatal ("print_x: unhandled type %d", cmd->in->args[i]);
	break;
      }

  va_end (ap);

  return (void *) self;
}
