/* 
   Written by Pieter J. Schoenmakers <tiggr@es.ele.tue.nl>

   Copyright (C) 1996 Pieter J. Schoenmakers.
   All rights reserved.

   This file is part of TOM.  TOM is distributed under the terms of the
   TOM License, a copy of which can be found in the TOM distribution; see
   the file LICENSE.

   $Id: util.h,v 1.24 1996/08/15 21:03:25 tiggr Exp $  */

#ifndef _trt_util_h_include_once
#define _trt_util_h_include_once
#include <stdarg.h>
#include "trt.h"

/* Return the address of the extension variables of the extension
   identified by EXTENSION_ID, in the object SELF.  */
TRT_INLINE void *
trt_ext_address (tom_object self, int extension_id)
{
  int offset = self->isa->info.eot->offset[extension_id];

  /* Check for an attempt to access variables of an extension not present
     in the object SELF.  */
  if (!offset)
    {
      /* This could still be the main State extension.  */
      if (extension_id
	  != *self->isa->info.extensions->extensions[0]->eid_in_a_global)
	abort ();
    }

  return (char *) self + offset;
}

/* Return !0 iff the selector argument descriptors A and B denote the same
   argument types.  */
TRT_INLINE int
trt_selector_args_match (struct trtd_selector_args *a,
			 struct trtd_selector_args *b)
{
  int i;

  /* This test would suffice if there were no dynamic loading, since the
     arg type descriptors are uniqued by the resolver.

     XXX Maybe they should be uniqued also by the dynamic resolver, or
     maybe they should bear unique identifiers like the selectors do.
     Then a simple comparison (either of the pointer or of the identity)
     will allways suffice.  */
  if (a == b)
    return 1;

  if (a->num != b->num)
    return 0;

  for (i = 0; i < a->num; i++)
    if (a->args[i] != b->args[i])
      return 0;

  return 1;
}

/* Return !0 iff the two selectors denote the same selector.  */
TRT_INLINE int
trt_selectors_equal (selector a, selector b)
{
  return a == b || a->sel_id == b->sel_id;
}

/* Modify the range which starts at START with LENGTH elements (which is
   -1 to indicate the maximum possible number of elements) to fit a range
   which starts at 0 with NUM elements (which can not be -1).  */
#define MODIFY_RANGE(START, LENGTH, NUM)  \
  do									\
    {									\
      if (START < 0)							\
	fatal (__FUNCTION__ ": bad range (%d, %d) for size %d",		\
	       (int) START, (int) LENGTH, (int) NUM);			\
      if (LENGTH == -1)							\
	LENGTH = NUM - START;						\
      else if (START + LENGTH > NUM)					\
	LENGTH = NUM - START;						\
      if (LENGTH <= 0)							\
	LENGTH = 0;							\
    } while (0)

/* Return the selector whose name is LEN bytes, pointed to by S.  */
selector trt_selector_named (char *s, int len);

/* Return the size of the arguments, to be passed to __builtin_apply, for
   the selector SEL.  */
int trt_all_arguments_size (selector sel);

/* Return a new byte-string containing the first N bytes pointed to by S.  */
tom_object byte_string_with_string (char *s, int n);

/* Return a new byte-string containing the bytes from the
   (zero-terminated) C-string S.  */
tom_object byte_string_with_c_string (const char *s);

/* Return a malloced zero terminated string with the LEN characters at
   BUF.  */
char *c_string_with_length (char *buf, int len);

/* Return a malloced zero terminated string with the byte STRING's contents.  */
char *c_string_with_tom_string (tom_object string);

/* Return the address of the ivar named NAME in the object SELF.  The
   variable should have the TYPE.  */
void *trt_address_of_ivar (tom_object self, tom_object name,
			   enum trt_type_encoding type);

/* Return the name of the indicated type.  */
char *trt_type_name (enum trt_type_encoding type);

/* Indicate that the object SELF does not recognize the selector CMD.  */
void trt_unrecognized (tom_object self, selector cmd);

/* Issue a fatal error (in the runtime).  */
void fatal (char *fmt, ...);
void unimplemented (char *fmt, ...);
void vfatal (char *fmt, va_list ap);

/* Sensible memory allocation.  */
void *xmalloc (unsigned int n);
void *xcalloc (unsigned int n, unsigned int m);
void *xrealloc (void *p, unsigned int n);
void xfree (void *p);

#ifdef DEBUG
#define ASSERT(X)  \
  do if (!(X))								\
    fatal ("%s:%d: assertion failed: %s", __FILE__, __LINE__, #X);	\
  while (0)
#else
#define ASSERT(X) do ; while (0)
#endif

#define SIGNAL(OBJ, CLASS, MSG)  \
  ({									\
     tom_object c = TRT_SEND ((reference_imp), _mr_c_tom_Condition,	\
			      SEL (r_for_r_class_r_message_r),		\
			      (OBJ), (CLASS), (MSG));			\
     TRT_SEND ((reference_imp), c, SEL (r_signal));})

#define RAISE(OBJ, CLASS, MSG)  \
  do									\
    {									\
      tom_object c = TRT_SEND ((reference_imp), _mr_c_tom_Condition,	\
			       SEL (r_for_r_class_r_message_r),		\
			       (OBJ), (CLASS), (MSG));			\
      TRT_SEND ((reference_imp), c, SEL (v_raise));			\
    } while (0)

#endif

#define PANIC_MODE_P()  \
  (c_tom_Runtime_panic_mode++)

#define PANIC_MODE_V()  \
  do									\
    {									\
      if (!c_tom_Runtime_quit_pending)					\
	c_tom_Runtime_panic_mode--;					\
      else								\
	TRT_SEND ((void_imp), _mr_c_tom_Runtime, SEL (v_panic_disable));  \
    } while (0)
